import pytest
import asyncio
import sys
from time import time

from a_sync.primitives.locks.semaphore import Semaphore
from tests.fixtures import TestSemaphore, increment


# ISSUES
# - We are unable to pass in an existing semaphore object, it attaches to a different loop.
#   Maybe a problem with test suite interaction?
# - semaphore modifier works fine with integer inputs


instance = TestSemaphore(1, sync=False)


def test_semaphore_init():
    assert Semaphore(1)._value == Semaphore()._value == 1


@increment
@pytest.mark.asyncio_cooperative
async def test_semaphore(i: int):
    start = time()
    assert await instance.test_fn() == 1
    duration = time() - start
    # There is a 1 second sleep in this fn. If the semaphore is not working, all tests will complete in 1 second.
    assert i < 3 or duration > i


@increment
@pytest.mark.asyncio_cooperative
async def test_semaphore_property(i: int):
    start = time()
    assert await instance.test_property == 2
    duration = time() - start
    # There is a 1 second sleep in this fn. If the semaphore is not working, all tests will complete in 1 second.
    assert i < 3 or duration > i


@increment
@pytest.mark.asyncio_cooperative
async def test_semaphore_cached_property(i: int):
    start = time()
    assert await instance.test_cached_property == 3
    duration = time() - start
    # There is a 1 second sleep in this fn but a semaphore override with a value of 50.
    # You can tell it worked correctly because the class-defined semaphore value is just one, whch would cause this test to fail if it were used.
    # If the override is not working, all tests will complete in just over 1 second.
    # We increased the threshold from 1.05 to 1.4 to help tests pass on slow github runners
    assert i == 1 or duration < 1.4


@pytest.mark.asyncio_cooperative
async def test_semaphore_acquire_release():
    semaphore = Semaphore(2)
    await semaphore.acquire()
    assert semaphore._value == 1
    semaphore.release()
    assert semaphore._value == 2


@pytest.mark.asyncio_cooperative
async def test_semaphore_blocking():
    semaphore = Semaphore(1)

    async def task():
        await semaphore.acquire()
        await asyncio.sleep(0.1)
        semaphore.release()

    await semaphore.acquire()
    task1 = asyncio.create_task(task())
    await asyncio.sleep(0.05)
    assert semaphore.locked()
    semaphore.release()
    await task1


@pytest.mark.asyncio_cooperative
async def test_semaphore_multiple_tasks():
    semaphore = Semaphore(2)
    results = []

    async def task(index):
        await semaphore.acquire()
        results.append(index)
        await asyncio.sleep(0.1)
        semaphore.release()

    tasks = [asyncio.create_task(task(i)) for i in range(4)]
    await asyncio.gather(*tasks)
    assert results == [0, 1, 2, 3]


@pytest.mark.asyncio_cooperative
async def test_semaphore_with_zero_initial_value():
    semaphore = Semaphore(0)

    async def task():
        await semaphore.acquire()
        return "done"

    task1 = asyncio.create_task(task())
    await asyncio.sleep(0.1)
    assert not task1.done()
    semaphore.release()
    result = await task1
    assert result == "done"


def test_semaphore_negative_initial_value():
    with pytest.raises(ValueError):
        Semaphore(-1)
    with pytest.raises(TypeError):
        Semaphore(None)
    with pytest.raises(TypeError):
        Semaphore("None")


@pytest.mark.asyncio_cooperative
async def test_semaphore_releasing_without_acquiring():
    semaphore = Semaphore(1)
    semaphore.release()
    assert semaphore._value == 2


"""
@pytest.mark.asyncio_cooperative
async def test_semaphore_with_custom_loop():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    semaphore = Semaphore(1)

    async def task():
        await semaphore.acquire()
        return 'done'

    task1 = loop.create_task(task())
    loop.call_soon(semaphore.release)
    result = loop.run_until_complete(task1)
    assert result == 'done'
    loop.close()
"""


@pytest.mark.asyncio_cooperative
async def test_concurrent_acquire_release():
    semaphore = Semaphore(2)
    results = []

    async def task(index):
        await semaphore.acquire()
        results.append(index)
        await asyncio.sleep(0.1)
        semaphore.release()

    tasks = [asyncio.create_task(task(i)) for i in range(10)]
    await asyncio.gather(*tasks)
    assert sorted(results) == list(range(10))


@pytest.mark.asyncio_cooperative
async def test_semaphore_max_integer_value():
    semaphore = Semaphore(sys.maxsize)
    await semaphore.acquire()
    assert semaphore._value == sys.maxsize - 1
    semaphore.release()
    assert semaphore._value == sys.maxsize


@pytest.mark.asyncio_cooperative
async def test_rapid_acquire_release():
    semaphore = Semaphore(1)

    async def task():
        for _ in range(100):
            await semaphore.acquire()
            semaphore.release()

    await asyncio.gather(*[task() for _ in range(10)])


@pytest.mark.asyncio_cooperative
async def test_exception_handling_during_acquire():
    semaphore = Semaphore(1)

    async def task():
        try:
            await semaphore.acquire()
            raise ValueError("Intentional error")
        except ValueError:
            semaphore.release()

    await task()
    assert semaphore._value == 1


@pytest.mark.asyncio_cooperative
async def test_cancelled_acquire():
    semaphore = Semaphore(1)

    async def task():
        await semaphore.acquire()
        await asyncio.sleep(0.5)
        semaphore.release()

    task1 = asyncio.create_task(task())
    await asyncio.sleep(0.1)

    async def waiting_task():
        await semaphore.acquire()

    waiting_task1 = asyncio.create_task(waiting_task())
    await asyncio.sleep(0.1)
    waiting_task1.cancel()
    await task1


@pytest.mark.asyncio_cooperative
async def test_delayed_release():
    semaphore = Semaphore(1)

    async def task():
        await semaphore.acquire()
        await asyncio.sleep(1)
        semaphore.release()

    task1 = asyncio.create_task(task())
    await asyncio.sleep(0.1)
    assert semaphore.locked()
    await task1


@pytest.mark.asyncio_cooperative
async def test_invalid_release():
    semaphore = Semaphore(1)
    semaphore.release()
    semaphore.release()
    assert semaphore._value == 3


@pytest.mark.asyncio_cooperative
async def test_custom_error_handling():
    semaphore = Semaphore(1)

    async def task():
        try:
            await semaphore.acquire()
            raise RuntimeError("Custom error")
        except RuntimeError:
            semaphore.release()

    await task()
    assert semaphore._value == 1


@pytest.mark.asyncio_cooperative
async def test_dynamic_resizing():
    semaphore = Semaphore(2)
    semaphore._value = 5
    await semaphore.acquire()
    assert semaphore._value == 4
    semaphore.release()
    assert semaphore._value == 5


@pytest.mark.asyncio_cooperative
async def test_high_frequency_operations():
    semaphore = Semaphore(1)

    async def task():
        for _ in range(1000):
            await semaphore.acquire()
            semaphore.release()

    await asyncio.gather(*[task() for _ in range(10)])


@pytest.mark.asyncio_cooperative
async def test_semaphore_as_context_manager():
    semaphore = Semaphore(1)

    async def task():
        async with semaphore:
            assert semaphore._value == 0

    await task()
    assert semaphore._value == 1


@pytest.mark.asyncio_cooperative
async def test_exception_in_release():
    semaphore = Semaphore(1)

    async def task():
        await semaphore.acquire()
        try:
            raise RuntimeError("Error during release")
        finally:
            semaphore.release()

    with pytest.raises(RuntimeError):
        await task()

    assert semaphore._value == 1


"""
@pytest.mark.asyncio_cooperative
async def test_external_interruptions():
    semaphore = Semaphore(1)

    async def task():
        await semaphore.acquire()
        await asyncio.sleep(0.1)
        semaphore.release()

    task1 = asyncio.create_task(task())
    await asyncio.sleep(0.05)
    task1.cancel()
    await asyncio.sleep(0.1)
    assert semaphore._value == 1
"""



from a_sync.primitives._debug cimport _DebugDaemonMixin

cdef class Semaphore(_DebugDaemonMixin):
    cdef str _name
    cdef unsigned long long __value
    cdef object _waiters
    cdef set _decorated
    cdef dict __dict__
    cpdef bint locked(self)
    cdef bint c_locked(self)
    cpdef object acquire(self)
    cdef object c_acquire(self)
    cpdef void release(self)
    cdef void c_release(self)
    cpdef void _wake_up_next(self)
    cdef void _c_wake_up_next(self)
            
cdef class DummySemaphore(Semaphore):
    pass

cdef class ThreadsafeSemaphore(Semaphore):
    cdef object semaphores, dummy
    cdef bint use_dummy
    cdef Semaphore c_get_semaphore(self)



from _typeshed import Incomplete
from a_sync.primitives._debug import _DebugDaemonMixin
from a_sync.primitives.locks import Event
from typing import Iterable, Optional

class CounterLock(_DebugDaemonMixin):
    """
    An async primitive that uses an internal counter to manage task synchronization.

    A coroutine can `await counter.wait_for(3)` and it will wait until the internal counter >= 3.
    If some other task executes `counter.value = 5` or `counter.set(5)`, the first coroutine will proceed as 5 >= 3.

    The internal counter can only be set to a value greater than the current value.

    See Also:
        :class:`CounterLockCluster` for managing multiple :class:`CounterLock` instances.
    """

    is_ready: Incomplete
    def __init__(self, start_value: int = 0, name: Optional[str] = None) -> None:
        """
        Initializes the :class:`CounterLock` with a starting value and an optional name.

        Args:
            start_value: The initial value of the counter.
            name: An optional name for the counter, used in debug logs.

        Examples:
            >>> counter = CounterLock(start_value=0, name="example_counter")
            >>> counter.value
            0
        """

    async def wait_for(self, value: int) -> bool:
        """
        Waits until the counter reaches or exceeds the specified value.

        This method will ensure the debug daemon is running if the counter is not ready.

        Args:
            value: The value to wait for.

        Examples:
            >>> counter = CounterLock(start_value=0)
            >>> await counter.wait_for(5)  # This will block until counter.value >= 5

        See Also:
            :meth:`CounterLock.set` to set the counter value.
        """

    def set(self, value: int) -> None:
        """
        Sets the counter to the specified value.

        This method internally uses the `value` property to enforce that the new value must be strictly greater than the current value.

        Args:
            value: The value to set the counter to. Must be strictly greater than the current value.

        Raises:
            ValueError: If the new value is less than or equal to the current value.

        Examples:
            >>> counter = CounterLock(start_value=0)
            >>> counter.set(5)
            >>> counter.value
            5

        See Also:
            :meth:`CounterLock.value` for direct value assignment.
        """

    @property
    def value(self) -> int:
        """
        Gets the current value of the counter.

        Examples:
            >>> counter = CounterLock(start_value=0)
            >>> counter.value
            0
        """

    @value.setter
    def value(self, value: int) -> None:
        """
        Sets the counter to a new value, waking up any waiters if the value increases beyond the value they are awaiting.

        Args:
            value: The new value of the counter.

        Raises:
            ValueError: If the new value is less than the current value.

        Examples:
            >>> counter = CounterLock(start_value=0)
            >>> counter.value = 5
            >>> counter.value
            5
            >>> counter.value = 3
            Traceback (most recent call last):
            ...
            ValueError: You cannot decrease the value.
        """

class CounterLockCluster:
    """
    An asyncio primitive that represents a collection of :class:`CounterLock` objects.

    `wait_for(i)` will wait until the value of all :class:`CounterLock` objects is >= i.

    See Also:
        :class:`CounterLock` for managing individual counters.
    """

    locks: Incomplete
    def __init__(self, counter_locks: Iterable[CounterLock]) -> None:
        """
        Initializes the :class:`CounterLockCluster` with a collection of :class:`CounterLock` objects.

        Args:
            counter_locks: The :class:`CounterLock` objects to manage.

        Examples:
            >>> lock1 = CounterLock(start_value=0)
            >>> lock2 = CounterLock(start_value=0)
            >>> cluster = CounterLockCluster([lock1, lock2])
        """

    async def wait_for(self, value: int) -> bool:
        """
        Waits until the value of all :class:`CounterLock` objects in the cluster reaches or exceeds the specified value.

        Args:
            value: The value to wait for.

        Examples:
            >>> lock1 = CounterLock(start_value=0)
            >>> lock2 = CounterLock(start_value=0)
            >>> cluster = CounterLockCluster([lock1, lock2])
            >>> await cluster.wait_for(5)  # This will block until all locks have value >= 5
        """



"""
This module extends Python's :func:`asyncio.as_completed` with additional functionality.
"""

from a_sync._typing import *

__all__ = ["as_completed"]

class tqdm_asyncio:
    @staticmethod
    def as_completed(*args, **kwargs) -> None: ...

# Names in __all__ with no definition:
#   as_completed



# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import asyncio

from a_sync import exceptions
from a_sync._typing import *
from a_sync.a_sync.modifiers.cache.memory import apply_async_memory_cache


class CacheArgs(TypedDict):
    """Typed dictionary for cache arguments."""

    cache_type: CacheType
    """The type of cache to use. Currently, only 'memory' is implemented."""

    cache_typed: bool
    """Whether to consider types for cache keys."""

    ram_cache_maxsize: Optional[int]
    """The maximum size for the LRU cache."""

    ram_cache_ttl: Optional[int]
    """The time-to-live for items in the LRU cache."""


@overload
def apply_async_cache(
    **modifiers: Unpack[CacheArgs],
) -> AsyncDecorator[P, T]:
    """Creates a decorator to apply an asynchronous cache.

    This overload is used when no coroutine function is provided. The returned
    decorator can be applied to a coroutine function later.

    Args:
        **modifiers: Cache configuration options such as `cache_type`, `cache_typed`, `ram_cache_maxsize`, and `ram_cache_ttl`.

    Examples:
        >>> @apply_async_cache(cache_type='memory', ram_cache_maxsize=100)
        ... async def my_function():
        ...     pass

    See Also:
        :func:`apply_async_memory_cache`
    """


@overload
def apply_async_cache(
    coro_fn: int,
    **modifiers: Unpack[CacheArgs],
) -> AsyncDecorator[P, T]:
    """Creates a decorator with `ram_cache_maxsize` set by an integer.

    This overload is used when an integer is provided as the first argument,
    which sets the `ram_cache_maxsize` for the cache. The returned decorator
    can be applied to a coroutine function later.

    Args:
        coro_fn: An integer to set as the max size for the cache.
        **modifiers: Additional cache configuration options.

    Examples:
        >>> @apply_async_cache(100, cache_type='memory')
        ... async def my_function():
        ...     pass

    See Also:
        :func:`apply_async_memory_cache`
    """


@overload
def apply_async_cache(
    coro_fn: CoroFn[P, T],
    **modifiers: Unpack[CacheArgs],
) -> CoroFn[P, T]:
    """Applies an asynchronous cache directly to a coroutine function.

    This overload is used when a coroutine function is provided. The cache is
    applied directly to the function.

    Args:
        coro_fn: The coroutine function to be cached.
        **modifiers: Cache configuration options such as `cache_type`, `cache_typed`, `ram_cache_maxsize`, and `ram_cache_ttl`.

    Examples:
        >>> async def my_function():
        ...     pass
        >>> cached_function = apply_async_cache(my_function, cache_type='memory', ram_cache_maxsize=100)

    See Also:
        :func:`apply_async_memory_cache`
    """


def apply_async_cache(
    coro_fn: Union[CoroFn[P, T], CacheType, int] = None,
    cache_type: CacheType = "memory",
    cache_typed: bool = False,
    ram_cache_maxsize: Optional[int] = None,
    ram_cache_ttl: Optional[int] = None,
) -> AsyncDecoratorOrCoroFn[P, T]:
    """Applies an asynchronous cache to a coroutine function.

    This function can be used to apply a cache to a coroutine function, either by passing the function directly or by using it as a decorator. The cache type is currently limited to 'memory'. If an unsupported cache type is specified, a `NotImplementedError` is raised.

    Args:
        coro_fn: The coroutine function to apply the cache to, or an integer to set as the max size.
        cache_type: The type of cache to use. Currently, only 'memory' is implemented.
        cache_typed: Whether to consider types for cache keys.
        ram_cache_maxsize: The maximum size for the LRU cache. If set to an integer, it overrides coro_fn.
        ram_cache_ttl: The time-to-live for items in the LRU cache.

    Raises:
        TypeError: If 'ram_cache_maxsize' is not an integer or None.
        FunctionNotAsync: If the provided function is not asynchronous.
        NotImplementedError: If an unsupported cache type is specified.

    Examples:
        >>> @apply_async_cache(cache_type='memory', ram_cache_maxsize=100)
        ... async def my_function():
        ...     pass

        >>> async def my_function():
        ...     pass
        >>> cached_function = apply_async_cache(my_function, cache_type='memory', ram_cache_maxsize=100)

        >>> @apply_async_cache(100, cache_type='memory')
        ... async def another_function():
        ...     pass

    See Also:
        :func:`apply_async_memory_cache`
    """

    # Parse Inputs
    if isinstance(coro_fn, int):
        assert ram_cache_maxsize is None
        ram_cache_maxsize = coro_fn
        coro_fn = None

    # Validate
    elif coro_fn is None:
        if ram_cache_maxsize is not None and not isinstance(ram_cache_maxsize, int):
            raise TypeError(
                "'lru_cache_maxsize' must be an integer or None.", ram_cache_maxsize
            )
    elif not asyncio.iscoroutinefunction(coro_fn):
        raise exceptions.FunctionNotAsync(coro_fn)

    if cache_type == "memory":
        cache_decorator = apply_async_memory_cache(
            maxsize=ram_cache_maxsize, ttl=ram_cache_ttl, typed=cache_typed
        )
        return cache_decorator if coro_fn is None else cache_decorator(coro_fn)
    elif cache_type == "disk":
        pass
    raise NotImplementedError(f"cache_type: {cache_type}")



import asyncio
import pytest

from a_sync import TaskMapping, create_task
from a_sync.task import _EmptySequenceError


async def _coro_fn(key):
    """Coroutine function for testing.

    Args:
        i: An integer input.

    Returns:
        A string representation of the incremented input.

    See Also:
        - :func:`a_sync.TaskMapping`
    """
    await asyncio.sleep(0.1)
    return str(key + 1) * (key + 1)


@pytest.mark.asyncio_cooperative
async def test_create_task():
    """Test the creation of an asynchronous task.

    Verifies that a task can be created using the `create_task`
    function with a coroutine and a specified name.

    See Also:
        - :func:`a_sync.create_task`
    """
    t = create_task(coro=asyncio.sleep(0), name="test")
    assert t.get_name() == "test", t
    await t


@pytest.mark.asyncio_cooperative
async def test_persistent_task():  # sourcery skip: simplify-boolean-comparison
    """Test the persistence of a task without a local reference.

    Checks if a task created without a local reference
    completes successfully by setting a nonlocal variable.
    The test ensures that the task completes by verifying
    the change in the nonlocal variable.

    See Also:
        - :func:`a_sync.create_task`
    """
    check = False

    async def task():
        await asyncio.sleep(1)
        nonlocal check
        check = True

    create_task(coro=task(), skip_gc_until_done=True)
    # there is no local reference to the newly created task. does it still complete?
    await asyncio.sleep(2)
    assert check is True


@pytest.mark.asyncio_cooperative
async def test_pruning():
    """Test task creation and handling without errors.

    Ensures that tasks can be created without causing errors.
    This test does not explicitly check for task pruning, despite
    its name, but rather focuses on task creation stability.

    See Also:
        - :func:`a_sync.create_task`
    """

    async def task():
        return

    create_task(coro=task(), skip_gc_until_done=True)
    await asyncio.sleep(0)
    # previously, it failed here
    create_task(coro=task(), skip_gc_until_done=True)


@pytest.mark.asyncio_cooperative
async def test_task_mapping_init():
    """Test initialization of TaskMapping.

    Verifies that the :class:`TaskMapping` class initializes correctly
    with the provided coroutine function and arguments. Checks
    the handling of function arguments and the task name.

    See Also:
        - :class:`a_sync.TaskMapping`
    """
    tasks = TaskMapping(_coro_fn)
    assert (
        tasks._wrapped_func is _coro_fn
    ), f"{tasks._wrapped_func} , {_coro_fn}, {tasks._wrapped_func == _coro_fn}"
    assert tasks._wrapped_func_kwargs == {}
    assert tasks._name is None
    tasks = TaskMapping(_coro_fn, name="test", kwarg0=1, kwarg1=None)
    assert tasks._wrapped_func_kwargs == {"kwarg0": 1, "kwarg1": None}
    assert tasks._name == "test"


@pytest.mark.asyncio_cooperative
async def test_task_mapping():
    """Test the functionality of TaskMapping.

    Checks the behavior of :class:`TaskMapping`, including task
    creation, retrieval, and execution. Verifies the ability
    to await the mapping and checks the return values of tasks.

    See Also:
        - :class:`a_sync.TaskMapping`
    """
    tasks = TaskMapping(_coro_fn)
    # does it return the correct type
    assert isinstance(tasks[0], asyncio.Task)
    # does it correctly return existing values
    assert tasks[1] is tasks[1]
    # does the task return the correct value
    assert await tasks[0] == "1"
    # can it do it again
    assert await tasks[0] == "1"
    # can we await the mapping?
    assert await tasks == {0: "1", 1: "22"}
    # can we await one from scratch?
    assert await TaskMapping(_coro_fn, range(5)) == {
        0: "1",
        1: "22",
        2: "333",
        3: "4444",
        4: "55555",
    }
    assert len(tasks) == 2


@pytest.mark.asyncio_cooperative
async def test_task_mapping_map_with_sync_iter():
    """Test TaskMapping with a synchronous iterator.

    Verifies that :class:`TaskMapping` can map over a synchronous
    iterator and correctly handle keys, values, and items.
    Ensures that mapping in progress raises a :class:`RuntimeError`
    when attempted concurrently.

    See Also:
        - :class:`a_sync.TaskMapping`
    """
    tasks = TaskMapping(_coro_fn)
    i = 0
    async for k, v in tasks.map(range(5)):
        assert isinstance(k, int)
        assert isinstance(v, str)
        if i < 4:
            # this shouldn't work since there is a mapping in progress
            with pytest.raises(RuntimeError):
                async for k in tasks.map(range(5)):
                    ...
        i += 1
    tasks = TaskMapping(_coro_fn)
    async for k in tasks.map(range(5), pop=False, yields="keys"):
        assert isinstance(k, int)

    # test keys
    for k in tasks.keys():
        assert isinstance(k, int)
    awaited = await tasks.keys()
    assert isinstance(awaited, list)
    for k in awaited:
        assert isinstance(k, int)
    async for k in tasks.keys():
        assert isinstance(k, int)

    # test values
    for v in tasks.values():
        assert isinstance(v, asyncio.Future)
        assert isinstance(await v, str)
    awaited = await tasks.values()
    assert isinstance(awaited, list)
    for v in awaited:
        assert isinstance(v, str)
    async for v in tasks.values():
        assert isinstance(v, str)

    # test items
    for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, asyncio.Future)
        assert isinstance(await v, str)
    awaited = await tasks.items()
    assert isinstance(awaited, list)
    for k, v in awaited:
        assert isinstance(k, int)
        assert isinstance(v, str)
    async for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, str)


@pytest.mark.asyncio_cooperative
async def test_task_mapping_map_with_async_iter():
    """Test TaskMapping with an asynchronous iterator.

    Verifies that :class:`TaskMapping` can map over an asynchronous
    iterator and correctly handle keys, values, and items.
    Ensures that mapping in progress raises a :class:`RuntimeError`
    when attempted concurrently.

    See Also:
        - :class:`a_sync.TaskMapping`
    """

    async def async_iter():
        for i in range(5):
            yield i

    tasks = TaskMapping(_coro_fn)
    i = 0
    async for k, v in tasks.map(async_iter()):
        assert isinstance(k, int)
        assert isinstance(v, str)
        if i < 4:
            # this shouldn't work since there is a mapping in progress
            with pytest.raises(RuntimeError):
                async for k in tasks.map(async_iter()):
                    ...
        i += 1
    tasks = TaskMapping(_coro_fn)
    async for k in tasks.map(async_iter(), pop=False, yields="keys"):
        assert isinstance(k, int)

    # test keys
    for k in tasks.keys():
        assert isinstance(k, int)
    awaited = await tasks.keys()
    assert isinstance(awaited, list)
    for k in awaited:
        assert isinstance(k, int)
    async for k in tasks.keys():
        assert isinstance(k, int)
    assert await tasks.keys().aiterbykeys() == list(range(5))
    assert await tasks.keys().aiterbyvalues() == list(range(5))
    assert await tasks.keys().aiterbykeys(reverse=True) == sorted(
        range(5), reverse=True
    )
    assert await tasks.keys().aiterbyvalues(reverse=True) == sorted(
        range(5), reverse=True
    )

    # test values
    for v in tasks.values():
        assert isinstance(v, asyncio.Future)
        assert isinstance(await v, str)
    awaited = await tasks.values()
    assert isinstance(awaited, list)
    for v in awaited:
        assert isinstance(v, str)
    async for v in tasks.values():
        assert isinstance(v, str)
    assert await tasks.values().aiterbykeys() == [str(i) * i for i in range(1, 6)]
    assert await tasks.values().aiterbyvalues() == [str(i) * i for i in range(1, 6)]
    assert await tasks.values().aiterbykeys(reverse=True) == [
        str(i) * i for i in sorted(range(1, 6), reverse=True)
    ]
    assert await tasks.values().aiterbyvalues(reverse=True) == [
        str(i) * i for i in sorted(range(1, 6), reverse=True)
    ]

    # test items
    for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, asyncio.Future)
        assert isinstance(await v, str)
    awaited = await tasks.items()
    assert isinstance(awaited, list)
    for k, v in awaited:
        assert isinstance(k, int)
        assert isinstance(v, str)
    async for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, str)
    assert await tasks.items().aiterbykeys() == [
        (i, str(i + 1) * (i + 1)) for i in range(5)
    ]
    assert await tasks.items().aiterbyvalues() == [
        (i, str(i + 1) * (i + 1)) for i in range(5)
    ]
    assert await tasks.items().aiterbykeys(reverse=True) == [
        (i, str(i + 1) * (i + 1)) for i in sorted(range(5), reverse=True)
    ]
    assert await tasks.items(pop=True).aiterbyvalues(reverse=True) == [
        (i, str(i + 1) * (i + 1)) for i in sorted(range(5), reverse=True)
    ]
    assert not tasks  # did pop work?


def test_taskmapping_views_sync():
    """Test synchronous views of TaskMapping.

    Checks the synchronous access to keys, values, and items
    in :class:`TaskMapping`. Verifies the state of these views before
    and after gathering tasks.

    See Also:
        - :class:`a_sync.TaskMapping`
    """
    tasks = TaskMapping(_coro_fn, range(5))

    # keys are currently empty until the loop has a chance to run
    _assert_len_dictviews(tasks, 0)
    tasks.gather()

    _assert_len_dictviews(tasks, 5)
    for k in tasks.keys():
        assert isinstance(k, int)

    # test values
    for v in tasks.values():
        assert isinstance(v, asyncio.Future)

    # test items
    for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, asyncio.Future)

    assert len(tasks.keys()) == 5
    for k in tasks.keys():
        assert isinstance(k, int)


@pytest.mark.asyncio_cooperative
async def test_task_mapping_empty_iterable():
    """Test TaskMapping with an empty iterable."""
    tasks = TaskMapping(_coro_fn, [])
    assert len(tasks) == 0
    with pytest.raises(_EmptySequenceError, match="\[\]"):
        await tasks


@pytest.mark.asyncio_cooperative
async def test_task_mapping_single_item():
    """Test TaskMapping with a single-item iterable."""
    tasks = TaskMapping(_coro_fn, [0])
    await asyncio.sleep(0.01)
    assert len(tasks) == 1
    assert await tasks == {0: "1"}


@pytest.mark.asyncio_cooperative
async def test_task_mapping_error_handling():
    """Test TaskMapping with a function that raises an exception."""

    async def error_fn(key):
        raise ValueError("Intentional error")

    tasks = TaskMapping(error_fn, [0])
    with pytest.raises(ValueError):
        await tasks


@pytest.mark.asyncio_cooperative
async def test_task_mapping_concurrency():
    """Test TaskMapping with a concurrency limit."""
    tasks = TaskMapping(_coro_fn, range(10), concurrency=3)
    running_tasks = [task for task in tasks.values() if not task.done()]
    assert len(running_tasks) <= 3


def _assert_len_dictviews(tasks, i):
    assert len(tasks.keys()) == i
    assert len(tasks.values()) == i
    assert len(tasks.items()) == i



from typed_envs import EnvVarFactory

envs = EnvVarFactory("EZASYNC")

# We have some envs here to help you debug your custom class implementations

DEBUG_CLASS_NAME = envs.create_env("DEBUG_CLASS_NAME", str, default="", verbose=False)
"""str: The name of the class to debug.

If you're only interested in debugging a specific class, set this to the class name.

Examples:
    To debug a class named `MyClass`, set the environment variable:
    
    .. code-block:: bash

        export EZASYNC_DEBUG_CLASS_NAME=MyClass

See Also:
    :func:`DEBUG_MODE` for enabling debug mode on all classes.
"""

DEBUG_MODE = envs.create_env(
    "DEBUG_MODE", bool, default=bool(DEBUG_CLASS_NAME), verbose=False
)
"""bool: Enables debug mode on all classes.

Set this environment variable to `True` to enable debug mode on all classes. 
If `DEBUG_CLASS_NAME` is set to a non-empty string, 
`DEBUG_MODE` will default to `True`.

Examples:
    To enable debug mode globally, set the environment variable:

    .. code-block:: bash

        export EZASYNC_DEBUG_MODE=True

    If you have set `DEBUG_CLASS_NAME` to a specific class, `DEBUG_MODE` will 
    automatically be `True` unless `DEBUG_CLASS_NAME` is an empty string.

See Also:
    :func:`DEBUG_CLASS_NAME` for debugging a specific class.
"""



"""
This module extends Python's :func:`asyncio.as_completed` with additional functionality.
"""

import asyncio

try:
    from tqdm.asyncio import tqdm_asyncio
except ImportError as e:

    class tqdm_asyncio:  # type: ignore [no-redef]
        @staticmethod
        def as_completed(*args, **kwargs):
            raise ImportError("You must have tqdm installed to use this feature")


from a_sync._typing import *
from a_sync.iter import ASyncIterator


@overload
def as_completed(
    fs: Iterable[Awaitable[T]],
    *,
    timeout: Optional[float] = None,
    return_exceptions: bool = False,
    aiter: Literal[False] = False,
    tqdm: bool = False,
    **tqdm_kwargs: Any
) -> Iterator[Coroutine[Any, Any, T]]: ...
@overload
def as_completed(
    fs: Iterable[Awaitable[T]],
    *,
    timeout: Optional[float] = None,
    return_exceptions: bool = False,
    aiter: Literal[True] = True,
    tqdm: bool = False,
    **tqdm_kwargs: Any
) -> ASyncIterator[T]: ...
@overload
def as_completed(
    fs: Mapping[K, Awaitable[V]],
    *,
    timeout: Optional[float] = None,
    return_exceptions: bool = False,
    aiter: Literal[False] = False,
    tqdm: bool = False,
    **tqdm_kwargs: Any
) -> Iterator[Coroutine[Any, Any, Tuple[K, V]]]: ...
@overload
def as_completed(
    fs: Mapping[K, Awaitable[V]],
    *,
    timeout: Optional[float] = None,
    return_exceptions: bool = False,
    aiter: Literal[True] = True,
    tqdm: bool = False,
    **tqdm_kwargs: Any
) -> ASyncIterator[Tuple[K, V]]: ...
def as_completed(
    fs,
    *,
    timeout: Optional[float] = None,
    return_exceptions: bool = False,
    aiter: bool = False,
    tqdm: bool = False,
    **tqdm_kwargs: Any
):
    """
    Concurrently awaits a list of awaitable objects or mappings of awaitables and returns an iterator of results.

    This function extends Python's :func:`asyncio.as_completed`, providing additional features for mixed use cases of individual awaitable objects and mappings of awaitables.

    Differences from :func:`asyncio.as_completed`:
    - Uses type hints for use with static type checkers.
    - Supports either individual awaitables or a k:v mapping of awaitables.
    - Can be used as an async iterator which yields the result values using :class:`ASyncIterator`.
    - Provides progress reporting using :mod:`tqdm` if 'tqdm' is set to True.

    Note:
        The `return_exceptions` parameter is used to wrap awaitables with exceptions if set to True, allowing exceptions to be returned as results instead of being raised.

    Args:
        fs: The awaitables to await concurrently. It can be a list of individual awaitables or a mapping of awaitables.
        timeout: The maximum time, in seconds, to wait for the completion of awaitables. Defaults to None (no timeout).
        return_exceptions: If True, exceptions are wrapped and returned as results instead of raising them. Defaults to False.
        aiter: If True, returns an async iterator of results using :class:`ASyncIterator`. Defaults to False.
        tqdm: If True, enables progress reporting using :mod:`tqdm`. Defaults to False.
        **tqdm_kwargs: Additional keyword arguments for :mod:`tqdm` if progress reporting is enabled.

    Examples:
        Awaiting individual awaitables:

        >>> awaitables = [async_function1(), async_function2()]
        >>> for coro in as_completed(awaitables):
        ...     val = await coro
        ...     ...

        >>> async for val in as_completed(awaitables, aiter=True):
        ...     ...

        Awaiting mappings of awaitables:

        >>> mapping = {'key1': async_function1(), 'key2': async_function2()}
        >>> for coro in as_completed(mapping):
        ...     k, v = await coro
        ...     ...

        >>> async for k, v in as_completed(mapping, aiter=True):
        ...     ...

    See Also:
        - :func:`asyncio.as_completed`
        - :class:`ASyncIterator`
    """
    if isinstance(fs, Mapping):
        return as_completed_mapping(
            mapping=fs,
            timeout=timeout or 0,
            return_exceptions=return_exceptions,
            aiter=aiter,
            tqdm=tqdm,
            tqdm_kwargs=tqdm_kwargs,
        )
    if return_exceptions:
        fs = [_exc_wrap(f) for f in fs]
    return (
        ASyncIterator(
            __yield_as_completed(fs, timeout=timeout, tqdm=tqdm, **tqdm_kwargs)
        )
        if aiter
        else (
            tqdm_asyncio.as_completed(fs, timeout=timeout, **tqdm_kwargs)
            if tqdm
            else asyncio.as_completed(fs, timeout=timeout)
        )
    )


cdef object as_completed_mapping(
    object mapping,
    unsigned int timeout,
    bint return_exceptions,
    bint aiter,
    bint tqdm,
    dict[str, object] tqdm_kwargs,
):
    """
    Concurrently awaits a mapping of awaitable objects and returns an iterator or async iterator of results.

    This function is designed to await a mapping of awaitable objects, where each key-value pair represents a unique awaitable. It enables concurrent execution and gathers results into an iterator or an async iterator.

    Args:
        mapping: A dictionary-like object where keys are of type K and values are awaitable objects of type V.
        timeout: The maximum time, in seconds, to wait for the completion of awaitables. Defaults to None (no timeout).
        return_exceptions: If True, exceptions are wrapped and returned as results instead of raising them. Defaults to False.
        aiter: If True, returns an async iterator of results using :class:`ASyncIterator`. Defaults to False.
        tqdm: If True, enables progress reporting using :mod:`tqdm`. Defaults to False.
        **tqdm_kwargs: Additional keyword arguments for :mod:`tqdm` if progress reporting is enabled.

    Example:
        >>> mapping = {'key1': async_function1(), 'key2': async_function2()}
        >>> for coro in as_completed_mapping(mapping):
        ...     k, v = await coro
        ...     ...

        >>> async for k, v in as_completed_mapping(mapping, aiter=True):
        ...     ...

    See Also:
        - :func:`as_completed`
        - :class:`ASyncIterator`
    """
    return as_completed(
        [
            __mapping_wrap(k, v, return_exceptions=return_exceptions)
            for k, v in mapping.items()
        ],
        timeout=timeout or None,
        aiter=aiter,
        tqdm=tqdm,
        **tqdm_kwargs
    )


async def _exc_wrap(awaitable: Awaitable[T]) -> Union[T, Exception]:
    """Wraps an awaitable to catch exceptions and return them instead of raising.

    Args:
        awaitable: The awaitable to wrap.

    Returns:
        The result of the awaitable or the exception if one is raised.
    """
    try:
        return await awaitable
    except Exception as e:
        return e


async def __yield_as_completed(
    futs: Iterable[Awaitable[T]],
    *,
    timeout: Optional[float] = None,
    return_exceptions: bool = False,
    tqdm: bool = False,
    **tqdm_kwargs: Any
) -> AsyncIterator[T]:
    """Yields results from awaitables as they complete.

    Args:
        futs: The awaitables to await.
        timeout: The maximum time, in seconds, to wait for the completion of awaitables. Defaults to None (no timeout).
        return_exceptions: If True, exceptions are wrapped and returned as results instead of raising them. Defaults to False.
        tqdm: If True, enables progress reporting using :mod:`tqdm`. Defaults to False.
        **tqdm_kwargs: Additional keyword arguments for :mod:`tqdm` if progress reporting is enabled.
    """
    for fut in as_completed(
        futs,
        timeout=timeout,
        return_exceptions=return_exceptions,
        tqdm=tqdm,
        **tqdm_kwargs
    ):
        yield await fut


@overload
async def __mapping_wrap(
    k: K, v: Awaitable[V], return_exceptions: Literal[True] = True
) -> Union[V, Exception]: ...
@overload
async def __mapping_wrap(
    k: K, v: Awaitable[V], return_exceptions: Literal[False] = False
) -> V: ...
async def __mapping_wrap(
    k: K, v: Awaitable[V], return_exceptions: bool = False
) -> Union[V, Exception]:
    """Wraps a key-value pair of awaitable to catch exceptions and return them with the key.

    Args:
        k: The key associated with the awaitable.
        v: The awaitable to wrap.
        return_exceptions: If True, exceptions are wrapped and returned as results instead of raising them. Defaults to False.

    Returns:
        A tuple of the key and the result of the awaitable or the exception if one is raised.
    """
    try:
        return k, await v
    except Exception as e:
        if return_exceptions:
            return k, e
        raise



# type: ignore [var-annotated]

"""
The `future.py` module provides functionality for handling asynchronous futures, 
including a decorator for converting callables into `ASyncFuture` objects and 
utilities for managing asynchronous computations.

Functions:
    future(callable: Union[Callable[P, Awaitable[T]], Callable[P, T]] = None, **kwargs: Unpack[ModifierKwargs]) -> Callable[P, Union[T, "ASyncFuture[T]"]]: 
        A decorator to convert a callable into an `ASyncFuture`, with optional modifiers.
    _gather_check_and_materialize(*things: Unpack[MaybeAwaitable[T]]) -> List[T]: 
        Gathers and materializes a list of awaitable or non-awaitable items.
    _check_and_materialize(thing: T) -> T: 
        Checks if an item is awaitable and materializes it.
    _materialize(meta: "ASyncFuture[T]") -> T: 
        Materializes the result of an `ASyncFuture`.

Classes:
    ASyncFuture: Represents an asynchronous future result.

TODO include a simple mathematics example a one complex example with numerous variables and operations
TODO include attribute access examples
TODO describe a bit more about both of the above 2 TODOs somewhere in this module-level docstring
TODO describe why a user would want to use these (to write cleaner code that doesn't require as many ugly gathers)
TODO include comparisons between the 'new way' with this future class and the 'old way' with gathers
"""

import asyncio
import concurrent.futures
from functools import partial, wraps
from inspect import isawaitable

import a_sync.asyncio
from a_sync._typing import *


def future(
    callable: AnyFn[P, T] = None,
    **kwargs: Unpack[ModifierKwargs],
) -> Callable[P, Union[T, "ASyncFuture[T]"]]:
    """
    A decorator function to convert a callable into an `ASyncFuture`, with optional modifiers.

    This function wraps the provided callable in an `_ASyncFutureWrappedFn` instance,
    which returns an `ASyncFuture` when called. The `ASyncFuture` allows the result
    of the callable to be awaited or accessed synchronously.

    Args:
        callable: The callable to convert. Defaults to None.
        **kwargs: Additional keyword arguments for the modifier.

    Returns:
        A callable that returns either the result or an `ASyncFuture`.

    Example:
        >>> @future
        ... async def async_function():
        ...     return 42
        >>> result = async_function()
        >>> isinstance(result, ASyncFuture)
        True

    See Also:
        :class:`ASyncFuture`
    """
    return _ASyncFutureWrappedFn(callable, **kwargs)


async def _gather_check_and_materialize(*things: Unpack[MaybeAwaitable[T]]) -> List[T]:
    """
    Gathers and materializes a list of awaitable or non-awaitable items.

    This function takes a list of items that may be awaitable or not, and
    returns a list of their results. Awaitable items are awaited, while
    non-awaitable items are returned as-is.

    Args:
        *things: Items to gather and materialize.

    Example:
        >>> async def async_fn(x):
        ...     return x
        >>> await _gather_check_and_materialize(async_fn(1), 2, async_fn(3))
        [1, 2, 3]
    """
    return await asyncio.gather(*[_check_and_materialize(thing) for thing in things])


async def _check_and_materialize(thing: T) -> T:
    """
    Checks if an item is awaitable and materializes it.

    If the item is awaitable, it is awaited and the result is returned.
    Otherwise, the item is returned as-is.

    Args:
        thing: The item to check and materialize.

    Example:
        >>> async def async_fn():
        ...     return 42
        >>> await _check_and_materialize(async_fn())
        42
    """
    return await thing if isawaitable(thing) else thing


def _materialize(meta: "ASyncFuture[T]") -> T:
    """
    Materializes the result of an `ASyncFuture`.

    This function attempts to run the event loop until the `ASyncFuture` is complete.
    If the event loop is already running, it raises a RuntimeError.

    Args:
        meta: The `ASyncFuture` to materialize.

    Raises:
        RuntimeError: If the event loop is running and the result cannot be awaited.

    Example:
        >>> future = ASyncFuture(asyncio.sleep(1, result=42))
        >>> _materialize(future)
        42

    See Also:
        :class:`ASyncFuture`
    """
    try:
        return asyncio.get_event_loop().run_until_complete(meta)
    except RuntimeError as e:
        raise RuntimeError(
            f"{meta} result is not set and the event loop is running, you will need to await it first"
        ) from e


MetaNumeric = Union[
    Numeric, "ASyncFuture[int]", "ASyncFuture[float]", "ASyncFuture[Decimal]"
]


@final
class ASyncFuture(concurrent.futures.Future, Awaitable[T]):
    """
    A class representing an asynchronous future result.

    Inherits from both `concurrent.futures.Future` and `Awaitable[T]`, allowing it to be used in both synchronous and asynchronous contexts.

    The `ASyncFuture` class provides additional functionality for arithmetic operations,
    comparisons, and conversions, making it versatile for various use cases.

    Example:
        >>> async def async_fn():
        ...     return 42
        >>> future = ASyncFuture(async_fn())
        >>> await future
        42

    Note:
        Arithmetic operations are implemented, allowing for mathematical operations on future results.
        You no longer have to choose between optimized async code and clean, readable code.

    Example:
        >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
        >>> future2 = ASyncFuture(asyncio.sleep(1, result=5))
        >>> future3 = ASyncFuture(asyncio.sleep(1, result=10))
        >>> future4 = ASyncFuture(asyncio.sleep(1, result=2))
        >>> result = (future1 + future2) / future3 ** future4
        >>> await result
        0.15

    Attribute Access:
        The `ASyncFuture` allows attribute access on the materialized result.

    Example:
        >>> class Example:
        ...     def __init__(self, value):
        ...         self.value = value
        >>> future = ASyncFuture(asyncio.sleep(1, result=Example(42)))
        >>> future.value
        42

    See Also:
        :func:`future` for creating `ASyncFuture` instances.
    """

    __slots__ = "__awaitable__", "__dependencies", "__dependants", "__task"

    def __init__(
        self, awaitable: Awaitable[T], dependencies: List["ASyncFuture"] = []
    ) -> None:  # sourcery skip: default-mutable-arg
        """
        Initializes an `ASyncFuture` with an awaitable and optional dependencies.

        Args:
            awaitable: The awaitable object.
            dependencies: A list of dependencies. Defaults to [].

        Example:
            >>> async def async_fn():
            ...     return 42
            >>> future = ASyncFuture(async_fn())
            >>> await future
            42
        """

        self.__awaitable__ = awaitable
        """The awaitable object."""

        self.__dependencies = dependencies
        """A list of dependencies."""

        for dependency in dependencies:
            assert isinstance(dependency, ASyncFuture)
            dependency.__dependants.append(self)

        self.__dependants: List[ASyncFuture] = []
        """A list of dependants."""

        self.__task = None
        """The task associated with the awaitable."""

        super().__init__()

    def __hash__(self) -> int:
        return hash(self.__awaitable__)

    def __repr__(self) -> str:
        string = f"<{self.__class__.__name__} {self._state} for {self.__awaitable__}"
        if self.cancelled():
            pass
        elif self.done():
            string += (
                f" exception={self.exception()}"
                if self.exception()
                else f" result={super().result()}"
            )
        return f"{string}>"

    def __list_dependencies(self, other) -> List["ASyncFuture"]:
        """
        Lists dependencies for the `ASyncFuture`.

        Args:
            other: The other dependency to list.

        Returns:
            A list of dependencies including the current and other `ASyncFuture`.
        """
        return [self, other] if isinstance(other, ASyncFuture) else [self]

    @property
    def result(self) -> Union[Callable[[], T], Any]:
        # sourcery skip: assign-if-exp, reintroduce-else
        """
        If this future is not done, it will work like `cf.Future.result`. It will block, await the awaitable, and return the result when ready.
        If this future is done and the result has attribute `result`, will return `getattr(future_result, 'result')`
        If this future is done and the result does NOT have attribute `result`, will again work like `cf.Future.result`

        Example:
            >>> future = ASyncFuture(asyncio.sleep(1, result=42))
            >>> future.result()
            42
        """
        if self.done():
            if hasattr(r := super().result(), "result"):
                # can be property, method, whatever. should work.
                return r.result
            # the result should be callable like an asyncio.Future
            return super().result
        return lambda: _materialize(self)

    def __getattr__(self, attr: str) -> Any:
        """
        Allows access to attributes of the materialized result.

        Args:
            attr: The attribute name to access.

        Returns:
            The attribute value from the materialized result.

        Example:
            >>> class Example:
            ...     def __init__(self, value):
            ...         self.value = value
            >>> future = ASyncFuture(asyncio.sleep(1, result=Example(42)))
            >>> future.value
            42
        """
        return getattr(_materialize(self), attr)

    def __getitem__(self, key) -> Any:
        """
        Allows item access on the materialized result.

        Args:
            key: The key to access.

        Returns:
            The item from the materialized result.

        Example:
            >>> future = ASyncFuture(asyncio.sleep(1, result={'key': 'value'}))
            >>> future['key']
            'value'
        """
        return _materialize(self)[key]

    # NOTE: broken, do not use. I think
    def __setitem__(self, key, value) -> None:
        """
        Allows setting an item on the materialized result.

        Args:
            key: The key to set.
            value: The value to set.

        Example:
            >>> future = ASyncFuture(asyncio.sleep(1, result={'key': 'value'}))
            >>> future['key'] = 'new_value'
        """
        _materialize(self)[key] = value

    # not sure what to call these
    def __contains__(self, key: Any) -> bool:
        """
        Checks if a key is in the materialized result.

        Args:
            key: The key to check.

        Returns:
            True if the key is in the materialized result, False otherwise.

        Example:
            >>> future = ASyncFuture(asyncio.sleep(1, result={'key': 'value'}))
            >>> 'key' in future
            True
        """
        return _materialize(
            ASyncFuture(
                self.__contains(key), dependencies=self.__list_dependencies(key)
            )
        )

    def __await__(self) -> Generator[Any, None, T]:
        """
        Makes the `ASyncFuture` awaitable.

        Example:
            >>> future = ASyncFuture(asyncio.sleep(1, result=42))
            >>> await future
            42
        """
        return self.__await().__await__()

    async def __await(self) -> T:
        if not self.done():
            self.set_result(await self.__task__)
        return self._result

    @property
    def __task__(self) -> "asyncio.Task[T]":
        """
        Returns the asyncio task associated with the awaitable, creating it if necessary.

        Example:
            >>> future = ASyncFuture(asyncio.sleep(1, result=42))
            >>> task = future.__task__
            >>> isinstance(task, asyncio.Task)
            True
        """
        if self.__task is None:
            self.__task = a_sync.asyncio.create_task(self.__awaitable__)
        return self.__task

    def __iter__(self):
        """
        Returns an iterator for the materialized result.

        Example:
            >>> future = ASyncFuture(asyncio.sleep(1, result=[1, 2, 3]))
            >>> for item in future:
            ...     print(item)
            1
            2
            3
        """
        return _materialize(self).__iter__()

    def __next__(self):
        """
        Returns the next item from the materialized result.

        Example:
            >>> future = ASyncFuture(asyncio.sleep(1, result=iter([1, 2, 3])))
            >>> next(future)
            1
        """
        return _materialize(self).__next__()

    def __enter__(self):
        """
        Enters the context of the materialized result.

        Example:
            >>> class SomeContext:
            ...     def __enter__(self):
            ...         return self
            ...     def __exit__(self, exc_type, exc_val, exc_tb):
            ...         pass
            >>> future = ASyncFuture(asyncio.sleep(1, result=SomeContext()))
            >>> with future as context:
            ...     context.do_something()
        """
        return _materialize(self).__enter__()

    def __exit__(self, *args):
        """
        Exits the context of the materialized result.

        Example:
            >>> class SomeContext:
            ...     def __enter__(self):
            ...         return self
            ...     def __exit__(self, exc_type, exc_val, exc_tb):
            ...         pass
            >>> future = ASyncFuture(asyncio.sleep(1, result=SomeContext()))
            >>> with future as context:
            ...     pass
            >>> # Context is exited here
        """
        return _materialize(self).__exit__(*args)

    @overload
    def __add__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...

    @overload
    def __add__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]": ...

    @overload
    def __add__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...

    @overload
    def __add__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...

    @overload
    def __add__(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __add__(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]": ...

    @overload
    def __add__(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]": ...

    @overload
    def __add__(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...

    @overload
    def __add__(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __add__(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __add__(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __add__(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __add__(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __add__(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    def __add__(self, other: MetaNumeric) -> "ASyncFuture":
        """
        Adds the result of this `ASyncFuture` to another value or `ASyncFuture`.

        Args:
            other: The value or `ASyncFuture` to add.

        Returns:
            A new `ASyncFuture` representing the sum.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=5))
            >>> result = future1 + future2
            >>> await result
            15
        """
        return ASyncFuture(
            self.__add(other), dependencies=self.__list_dependencies(other)
        )

    @overload
    def __sub__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...

    @overload
    def __sub__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]": ...

    @overload
    def __sub__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...

    @overload
    def __sub__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...

    @overload
    def __sub__(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __sub__(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]": ...

    @overload
    def __sub__(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]": ...

    @overload
    def __sub__(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...

    @overload
    def __sub__(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __sub__(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __sub__(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __sub__(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __sub__(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __sub__(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    def __sub__(self, other: MetaNumeric) -> "ASyncFuture":
        """
        Subtracts another value or `ASyncFuture` from the result of this `ASyncFuture`.

        Args:
            other: The value or `ASyncFuture` to subtract.

        Returns:
            A new `ASyncFuture` representing the difference.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=5))
            >>> result = future1 - future2
            >>> await result
            5
        """
        return ASyncFuture(
            self.__sub(other), dependencies=self.__list_dependencies(other)
        )

    def __mul__(self, other) -> "ASyncFuture":
        """
        Multiplies the result of this `ASyncFuture` by another value or `ASyncFuture`.

        Args:
            other: The value or `ASyncFuture` to multiply.

        Returns:
            A new `ASyncFuture` representing the product.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=5))
            >>> result = future1 * future2
            >>> await result
            50
        """
        return ASyncFuture(
            self.__mul(other), dependencies=self.__list_dependencies(other)
        )

    def __pow__(self, other) -> "ASyncFuture":
        """
        Raises the result of this `ASyncFuture` to the power of another value or `ASyncFuture`.

        Args:
            other: The exponent value or `ASyncFuture`.

        Returns:
            A new `ASyncFuture` representing the power.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=2))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=3))
            >>> result = future1 ** future2
            >>> await result
            8
        """
        return ASyncFuture(
            self.__pow(other), dependencies=self.__list_dependencies(other)
        )

    def __truediv__(self, other) -> "ASyncFuture":
        """
        Divides the result of this `ASyncFuture` by another value or `ASyncFuture`.

        Args:
            other: The divisor value or `ASyncFuture`.

        Returns:
            A new `ASyncFuture` representing the quotient.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=2))
            >>> result = future1 / future2
            >>> await result
            5.0
        """
        return ASyncFuture(
            self.__truediv(other), dependencies=self.__list_dependencies(other)
        )

    def __floordiv__(self, other) -> "ASyncFuture":
        """
        Performs floor division of the result of this `ASyncFuture` by another value or `ASyncFuture`.

        Args:
            other: The divisor value or `ASyncFuture`.

        Returns:
            A new `ASyncFuture` representing the floor division result.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=3))
            >>> result = future1 // future2
            >>> await result
            3
        """
        return ASyncFuture(
            self.__floordiv(other), dependencies=self.__list_dependencies(other)
        )

    def __pow__(self, other) -> "ASyncFuture":
        """
        Raises the result of this `ASyncFuture` to the power of another value or `ASyncFuture`.

        Args:
            other: The exponent value or `ASyncFuture`.

        Returns:
            A new `ASyncFuture` representing the power.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=2))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=3))
            >>> result = future1 ** future2
            >>> await result
            8
        """
        return ASyncFuture(
            self.__pow(other), dependencies=self.__list_dependencies(other)
        )

    @overload
    def __radd__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...

    @overload
    def __radd__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]": ...

    @overload
    def __radd__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...

    @overload
    def __radd__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __radd__(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    def __radd__(self, other) -> "ASyncFuture":
        """
        Adds another value or `ASyncFuture` to the result of this `ASyncFuture`.

        Args:
            other: The value or `ASyncFuture` to add.

        Returns:
            A new `ASyncFuture` representing the sum.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> result = 5 + future1
            >>> await result
            15
        """
        return ASyncFuture(
            self.__radd(other), dependencies=self.__list_dependencies(other)
        )

    @overload
    def __rsub__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...

    @overload
    def __rsub__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]": ...

    @overload
    def __rsub__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...

    @overload
    def __rsub__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    def __rsub__(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    def __rsub__(self, other) -> "ASyncFuture":
        """
        Subtracts the result of this `ASyncFuture` from another value or `ASyncFuture`.

        Args:
            other: The value or `ASyncFuture` to subtract from.

        Returns:
            A new `ASyncFuture` representing the difference.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> result = 20 - future1
            >>> await result
            10
        """
        return ASyncFuture(
            self.__rsub(other), dependencies=self.__list_dependencies(other)
        )

    def __rmul__(self, other) -> "ASyncFuture":
        """
        Multiplies another value or `ASyncFuture` by the result of this `ASyncFuture`.

        Args:
            other: The value or `ASyncFuture` to multiply.

        Returns:
            A new `ASyncFuture` representing the product.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> result = 2 * future1
            >>> await result
            20
        """
        return ASyncFuture(
            self.__rmul(other), dependencies=self.__list_dependencies(other)
        )

    def __rtruediv__(self, other) -> "ASyncFuture":
        """
        Divides another value or `ASyncFuture` by the result of this `ASyncFuture`.

        Args:
            other: The value or `ASyncFuture` to divide.

        Returns:
            A new `ASyncFuture` representing the quotient.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=2))
            >>> result = 10 / future1
            >>> await result
            5.0
        """
        return ASyncFuture(
            self.__rtruediv(other), dependencies=self.__list_dependencies(other)
        )

    def __rfloordiv__(self, other) -> "ASyncFuture":
        """
        Performs floor division of another value or `ASyncFuture` by the result of this `ASyncFuture`.

        Args:
            other: The value or `ASyncFuture` to divide.

        Returns:
            A new `ASyncFuture` representing the floor division result.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=3))
            >>> result = 10 // future1
            >>> await result
            3
        """
        return ASyncFuture(
            self.__rfloordiv(other), dependencies=self.__list_dependencies(other)
        )

    def __rpow__(self, other) -> "ASyncFuture":
        """
        Raises another value or `ASyncFuture` to the power of the result of this `ASyncFuture`.

        Args:
            other: The base value or `ASyncFuture`.

        Returns:
            A new `ASyncFuture` representing the power.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=3))
            >>> result = 2 ** future1
            >>> await result
            8
        """
        return ASyncFuture(
            self.__rpow(other), dependencies=self.__list_dependencies(other)
        )

    def __eq__(self, other) -> "ASyncFuture":
        """
        Compares the result of this `ASyncFuture` with another value or `ASyncFuture` for equality.

        Args:
            other: The value or `ASyncFuture` to compare.

        Returns:
            A new `ASyncFuture` representing the equality comparison result.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> result = future1 == future2
            >>> await result
            True
        """
        return bool(
            ASyncFuture(self.__eq(other), dependencies=self.__list_dependencies(other))
        )

    def __gt__(self, other) -> "ASyncFuture":
        """
        Compares the result of this `ASyncFuture` with another value or `ASyncFuture` for greater than.

        Args:
            other: The value or `ASyncFuture` to compare.

        Returns:
            A new `ASyncFuture` representing the greater than comparison result.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=5))
            >>> result = future1 > future2
            >>> await result
            True
        """
        return ASyncFuture(
            self.__gt(other), dependencies=self.__list_dependencies(other)
        )

    def __ge__(self, other) -> "ASyncFuture":
        """
        Compares the result of this `ASyncFuture` with another value or `ASyncFuture` for greater than or equal.

        Args:
            other: The value or `ASyncFuture` to compare.

        Returns:
            A new `ASyncFuture` representing the greater than or equal comparison result.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> result = future1 >= future2
            >>> await result
            True
        """
        return ASyncFuture(
            self.__ge(other), dependencies=self.__list_dependencies(other)
        )

    def __lt__(self, other) -> "ASyncFuture":
        """
        Compares the result of this `ASyncFuture` with another value or `ASyncFuture` for less than.

        Args:
            other: The value or `ASyncFuture` to compare.

        Returns:
            A new `ASyncFuture` representing the less than comparison result.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=5))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=10))
            >>> result = future1 < future2
            >>> await result
            True
        """
        return ASyncFuture(
            self.__lt(other), dependencies=self.__list_dependencies(other)
        )

    def __le__(self, other) -> "ASyncFuture":
        """
        Compares the result of this `ASyncFuture` with another value or `ASyncFuture` for less than or equal.

        Args:
            other: The value or `ASyncFuture` to compare.

        Returns:
            A new `ASyncFuture` representing the less than or equal comparison result.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=5))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=5))
            >>> result = future1 <= future2
            >>> await result
            True
        """
        return ASyncFuture(
            self.__le(other), dependencies=self.__list_dependencies(other)
        )

    # Maths

    @overload
    async def __add(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...

    @overload
    async def __add(
        self: "ASyncFuture[float]", other: float
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __add(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...

    @overload
    async def __add(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...

    @overload
    async def __add(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __add(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __add(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __add(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...

    @overload
    async def __add(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __add(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __add(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __add(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __add(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __add(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    async def __add(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a + b

    @overload
    async def __sub(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[float]", other: float
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __sub(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...

    @overload
    async def __sub(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __sub(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    async def __sub(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a - b

    async def __mul(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a * b

    async def __truediv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a / b

    async def __floordiv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a // b

    async def __pow(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a**b

    # rMaths
    @overload
    async def __radd(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[float]", other: float
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[float]", other: int
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: float
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    async def __radd(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a + b

    @overload
    async def __rsub(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[float]", other: float
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[float]", other: int
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: float
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...

    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...

    async def __rsub(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a - b

    async def __rmul(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a * b

    async def __rtruediv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a / b

    async def __rfloordiv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a // b

    async def __rpow(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a**b

    async def __iadd(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a + b
        return self._result

    async def __isub(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a - b
        return self._result

    async def __imul(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a * b
        return self._result

    async def __itruediv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a / b
        return self._result

    async def __ifloordiv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a // b
        return self._result

    async def __ipow(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a**b
        return self._result

    # Comparisons
    async def __eq(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a == b

    async def __gt(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a > b

    async def __ge(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a >= b

    async def __lt(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a < b

    async def __le(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a <= b

    # not sure what to call these
    async def __contains(self, item: Any) -> bool:
        _self, _item = await _gather_check_and_materialize(self, item)
        return _item in _self

    # conversion
    # NOTE: We aren't allowed to return ASyncFutures here :(
    def __bool__(self) -> bool:
        return bool(_materialize(self))

    def __bytes__(self) -> bytes:
        return bytes(_materialize(self))

    def __str__(self) -> str:
        return str(_materialize(self))

    def __int__(self) -> int:
        return int(_materialize(self))

    def __float__(self) -> float:
        return float(_materialize(self))

    # WIP internals

    @property
    def __dependants__(self) -> Set["ASyncFuture"]:
        """
        Returns the set of dependants for this `ASyncFuture`, including nested dependants.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=42))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=10), dependencies=[future1])
            >>> dependants = future1.__dependants__
            >>> future2 in dependants
            True
        """
        dependants = set()
        for dep in self.__dependants:
            dependants.add(dep)
            dependants.union(dep.__dependants__)
        return dependants

    @property
    def __dependencies__(self) -> Set["ASyncFuture"]:
        """
        Returns the set of dependencies for this `ASyncFuture`, including nested dependencies.

        Example:
            >>> future1 = ASyncFuture(asyncio.sleep(1, result=42))
            >>> future2 = ASyncFuture(asyncio.sleep(1, result=10), dependencies=[future1])
            >>> dependencies = future2.__dependencies__
            >>> future1 in dependencies
            True
        """
        dependencies = set()
        for dep in self.__dependencies:
            dependencies.add(dep)
            dependencies.union(dep.__dependencies__)
        return dependencies

    def __sizeof__(self) -> int:
        if isinstance(self.__awaitable__, Coroutine):
            return sum(
                sys.getsizeof(v) for v in self.__awaitable__.cr_frame.f_locals.values()
            )
        elif isinstance(self.__awaitable__, asyncio.Future):
            raise NotImplementedError
        raise NotImplementedError


@final
class _ASyncFutureWrappedFn(Callable[P, ASyncFuture[T]]):
    """
    A callable class to wrap functions and return `ASyncFuture` objects.

    This class is used internally by the `future` decorator to wrap a function
    and return an `ASyncFuture` when the function is called. It allows the
    function to be executed asynchronously and its result to be awaited.

    Example:
        >>> def sync_fn():
        ...     return 42
        >>> wrapped_fn = _ASyncFutureWrappedFn(sync_fn)
        >>> future = wrapped_fn()
        >>> isinstance(future, ASyncFuture)
        True

    Note:
        This is not part of the public API. Use the `future` decorator instead.
    """

    __slots__ = "callable", "wrapped", "_callable_name"

    def __init__(
        self,
        callable: AnyFn[P, T] = None,
        **kwargs: Unpack[ModifierKwargs],
    ):
        from a_sync import a_sync

        if callable:
            self.callable = callable
            """The callable function."""

            self._callable_name = callable.__name__
            """The name of the callable function."""

            a_sync_callable = a_sync(callable, default="async", **kwargs)

            @wraps(callable)
            def future_wrap(*args: P.args, **kwargs: P.kwargs) -> "ASyncFuture[T]":
                return ASyncFuture(a_sync_callable(*args, **kwargs, sync=False))

            self.wrapped = future_wrap
            """The wrapped function returning ASyncFuture."""
        else:
            self.wrapped = partial(_ASyncFutureWrappedFn, **kwargs)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ASyncFuture[T]:
        return self.wrapped(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.callable}>"

    def __get__(
        self, instance: I, owner: Type[I]
    ) -> Union[Self, "_ASyncFutureInstanceMethod[I, P, T]"]:
        return self if owner is None else _ASyncFutureInstanceMethod(self, instance)


@final
class _ASyncFutureInstanceMethod(Generic[I, P, T]):
    # NOTE: probably could just replace this with functools.partial
    """
    A class to handle instance methods wrapped as `ASyncFuture`.

    This class is used internally to manage instance methods that are wrapped
    by `_ASyncFutureWrappedFn`. It ensures that the method is bound to the
    instance and returns an `ASyncFuture` when called.

    Example:
        >>> class MyClass:
        ...     @_ASyncFutureWrappedFn
        ...     def method(self):
        ...         return 42
        >>> instance = MyClass()
        >>> future = instance.method()
        >>> isinstance(future, ASyncFuture)
        True

    Note:
        This is not part of the public API. Use the `future` decorator instead.
    """

    __module__: str
    """The module name of the wrapper."""

    __name__: str
    """The name of the wrapper."""

    __qualname__: str
    """The qualified name of the wrapper."""

    __doc__: Optional[str]
    """The docstring of the wrapper."""

    __annotations__: Dict[str, Any]
    """The annotations of the wrapper."""

    __instance: I
    """The instance to which the method is bound."""

    __wrapper: _ASyncFutureWrappedFn[P, T]
    """The wrapper function."""

    def __init__(
        self,
        wrapper: _ASyncFutureWrappedFn[P, T],
        instance: I,
    ) -> None:  # sourcery skip: use-contextlib-suppress
        try:
            self.__module__ = wrapper.__module__
        except AttributeError:
            pass
        try:
            self.__name__ = wrapper.__name__
        except AttributeError:
            pass
        try:
            self.__qualname__ = wrapper.__qualname__
        except AttributeError:
            pass
        try:
            self.__doc__ = wrapper.__doc__
        except AttributeError:
            pass
        try:
            self.__annotations__ = wrapper.__annotations__
        except AttributeError:
            pass
        try:
            self.__dict__.update(wrapper.__dict__)
        except AttributeError:
            pass
        self.__instance = instance
        self.__wrapper = wrapper

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} for {self.__wrapper.callable} bound to {self.__instance}>"

    def __call__(self, /, *fn_args: P.args, **fn_kwargs: P.kwargs) -> T:
        return self.__wrapper(self.__instance, *fn_args, **fn_kwargs)


__all__ = ["future", "ASyncFuture"]



"""
This module provides various semaphore implementations, including a debug-enabled semaphore,
a dummy semaphore that does nothing, and a threadsafe semaphore for use in multi-threaded applications.
"""

import asyncio
import functools
from collections import defaultdict, deque
from threading import Thread, current_thread
from typing import Container

from a_sync._typing import *
from a_sync.primitives._debug cimport _DebugDaemonMixin


async def __acquire() -> Literal[True]:
    return True


cdef class Semaphore(_DebugDaemonMixin):
    """
    A semaphore with additional debugging capabilities inherited from :class:`_DebugDaemonMixin`.

    This semaphore includes debug logging capabilities that are activated when the semaphore has waiters.
    It allows rewriting the pattern of acquiring a semaphore within a coroutine using a decorator.

    Example:
        You can write this pattern:

        ```
        semaphore = Semaphore(5)

        async def limited():
            async with semaphore:
                return 1
        ```

        like this:

        ```
        semaphore = Semaphore(5)

        @semaphore
        async def limited():
            return 1
        ```

    See Also:
        :class:`_DebugDaemonMixin` for more details on debugging capabilities.
    """

    def __cinit__(self) -> None:
        self.__waiters = deque()
        self._decorated: Set[str] = set()
    
    def __init__(self, value: int=1, name=None, loop=None, **kwargs) -> None:
        """
        Initialize the semaphore with a given value and optional name for debugging.

        Args:
            value: The initial value for the semaphore.
            name (optional): An optional name used only to provide useful context in debug logs.
        """
        super().__init__(loop=loop)
        if value < 0:
            raise ValueError("Semaphore initial value must be >= 0")

        try:
            self.__value = value
        except OverflowError as e:
            raise OverflowError(
                {"error": str(e), "value": value, "max value": 18446744073709551615},
                "If you need a Semaphore with a larger value, you should just use asyncio.Semaphore",
            ) from e.__cause__
            
        self._name = name or getattr(self, "__origin__", "")

    def __call__(self, fn: CoroFn[P, T]) -> CoroFn[P, T]:
        """
        Decorator method to wrap coroutine functions with the semaphore.

        This allows rewriting the pattern of acquiring a semaphore within a coroutine using a decorator.

        Example:
            semaphore = Semaphore(5)

            @semaphore
            async def limited():
                return 1
        """
        return self.decorate(fn)  # type: ignore [arg-type, return-value]

    def __repr__(self) -> str:
        representation = f"<{self.__class__.__name__} name={self._name} value={self.__value} waiters={len(self)}>"
        if self._decorated:
            representation = f"{representation[:-1]} decorates={self._decorated}"
        return representation

    async def __aenter__(self):
        coro = self.c_acquire()
        await coro
        # We have no use for the "as ..."  clause in the with
        # statement for locks.
        return None

    async def __aexit__(self, exc_type, exc, tb):
        self.c_release()

    cpdef bint locked(self):
        """Returns True if semaphore cannot be acquired immediately."""
        return self.c_locked()
    
    cdef bint c_locked(self):
        """Returns True if semaphore cannot be acquired immediately."""
        if self.__value == 0:
            return True
        waiters = self.__waiters 
        if any(not w.cancelled() for w in (waiters or ())):
            return True

    def __len__(self) -> int:
        return len(self.__waiters)

    def decorate(self, fn: CoroFn[P, T]) -> CoroFn[P, T]:
        """
        Wrap a coroutine function to ensure it runs with the semaphore.

        Example:
            semaphore = Semaphore(5)

            @semaphore
            async def limited():
                return 1
        """
        if not asyncio.iscoroutinefunction(fn):
            raise TypeError(f"{fn} must be a coroutine function")

        @functools.wraps(fn)
        async def semaphore_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            async with self:
                return await fn(*args, **kwargs)

        self._decorated.add(f"{fn.__module__}.{fn.__name__}")
        return semaphore_wrapper

    cpdef object acquire(self):
        """
        Acquire the semaphore, ensuring that debug logging is enabled if there are waiters.

        If the internal counter is larger than zero on entry, decrement it by one and return
        True immediately.  If it is zero on entry, block, waiting until some other coroutine 
        has called release() to make it larger than 0, and then return True.

        If the semaphore value is zero or less, the debug daemon is started to log the state of the semaphore.

        Returns:
            True when the semaphore is successfully acquired.
        """
        return self.c_acquire()
    
    cdef object c_acquire(self):
        """
        Acquire the semaphore, ensuring that debug logging is enabled if there are waiters.

        If the internal counter is larger than zero on entry, decrement it by one and return
        True immediately.  If it is zero on entry, block, waiting until some other coroutine 
        has called release() to make it larger than 0, and then return True.

        If the semaphore value is zero or less, the debug daemon is started to log the state of the semaphore.

        Returns:
            True when the semaphore is successfully acquired.
        """
        if self.__value <= 0:
            self._c_ensure_debug_daemon((),{})

        if not self.c_locked():
            self.__value -= 1
            return __acquire()

        return self.__acquire()

    async def __acquire(self) -> Literal[True]:
        # Finally block should be called before the CancelledError
        # handling as we don't want CancelledError to call
        # _wake_up_first() and attempt to wake up itself.
        
        cdef object fut = self._c_get_loop().create_future()
        self.__waiters.append(fut)
        try:
            try:
                await fut
            finally:
                self.__waiters.remove(fut)
        except asyncio.exceptions.CancelledError:
            if not fut.cancelled():
                self.__value += 1
                self._c_wake_up_next()
            raise

        if self.__value > 0:
            self._c_wake_up_next()

        return True

    cpdef void release(self):
        """Release a semaphore, incrementing the internal counter by one.

        When it was zero on entry and another coroutine is waiting for it to
        become larger than zero again, wake up that coroutine.
        """
        self.__value += 1
        self._c_wake_up_next()
    
    cdef void c_release(self):
        self.__value += 1
        self._c_wake_up_next()

    @property
    def name(self) -> str:
        return self._name

    @property
    def _value(self) -> int:
        # required for subclass compatability
        return self.__value
    
    @_value.setter
    def _value(self, unsigned long long value):
        # required for subclass compatability
        self.__value = value

    @property
    def _waiters(self) -> int:
        # required for subclass compatability
        return self.__waiters
    
    @_waiters.setter
    def _waiters(self, value: Container):
        # required for subclass compatability
        self.__waiters = value

    cpdef void _wake_up_next(self):
        """Wake up the first waiter that isn't done."""
        if not self.__waiters:
            return

        for fut in self.__waiters:
            if not fut.done():
                self.__value -= 1
                fut.set_result(True)
                return
    
    cdef void _c_wake_up_next(self):
        """Wake up the first waiter that isn't done."""
        if not self.__waiters:
            return

        for fut in self.__waiters:
            if not fut.done():
                self.__value -= 1
                fut.set_result(True)
                return
            
    async def _debug_daemon(self) -> None:
        """
        Daemon coroutine (runs in a background task) which will emit a debug log every minute while the semaphore has waiters.

        This method is part of the :class:`_DebugDaemonMixin` and is used to provide detailed logging information
        about the semaphore's state when it is being waited on.

        Example:
            semaphore = Semaphore(5)

            async def monitor():
                await semaphore._debug_daemon()
        """
        while self.__waiters:
            await asyncio.sleep(60)
            self.get_logger().debug(
                "%s has %s waiters for any of: %s",
                self,
                len(self),
                self._decorated,
            )


cdef class DummySemaphore(Semaphore):
    """
    A dummy semaphore that implements the standard :class:`asyncio.Semaphore` API but does nothing.

    This class is useful for scenarios where a semaphore interface is required but no actual synchronization is needed.

    Example:
        dummy_semaphore = DummySemaphore()

        async def no_op():
            async with dummy_semaphore:
                return 1
    """

    def __cinit__(self):
        self.__value = 0
        self.__waiters = deque()
        self._decorated = None

    def __init__(self, name: Optional[str] = None):
        """
        Initialize the dummy semaphore with an optional name.

        Args:
            name (optional): An optional name for the dummy semaphore.
        """
        self._name = name

    def __repr__(self) -> str:
        return "<{} name={}>".format(self.__class__.__name__, self._name)

    async def acquire(self) -> Literal[True]:
        """Acquire the dummy semaphore, which is a no-op."""
        return True
    
    async def c_acquire(self) -> Literal[True]:
        return True

    cpdef void release(self):
        """No-op release method."""
    
    cdef void c_release(self):
        """No-op release method."""

    async def __aenter__(self):
        """No-op context manager entry."""
        return self

    async def __aexit__(self, *args):
        """No-op context manager exit."""


cdef class ThreadsafeSemaphore(Semaphore):
    """
    A semaphore that works in a multi-threaded environment.

    This semaphore ensures that the program functions correctly even when used with multiple event loops.
    It provides a workaround for edge cases involving multiple threads and event loops by using a separate semaphore
    for each thread.

    Example:
        semaphore = ThreadsafeSemaphore(5)

        async def limited():
            async with semaphore:
                return 1

    See Also:
        :class:`Semaphore` for the base class implementation.
    """

    def __init__(self, value: Optional[int], name: Optional[str] = None) -> None:
        """
        Initialize the threadsafe semaphore with a given value and optional name.

        Args:
            value: The initial value for the semaphore, should be an integer.
            name (optional): An optional name for the semaphore.
        """
        assert isinstance(value, int), f"{value} should be an integer."
        super().__init__(value, name=name)
        
        self.use_dummy = value is -1
        if self.use_dummy:
            self.semaphores = {}
            self.dummy = DummySemaphore(name=name)
        else:
            self.semaphores: DefaultDict[Thread, Semaphore] = defaultdict(lambda: Semaphore(value, name=self._name))  # type: ignore [arg-type]
            self.dummy = None

    def __len__(self) -> int:
        return sum(len(sem._waiters) for sem in self.semaphores.values())

    @property
    def semaphore(self) -> Semaphore:
        """
        Returns the appropriate semaphore for the current thread.

        NOTE: We can't cache this property because we need to check the current thread every time we access it.

        Example:
            semaphore = ThreadsafeSemaphore(5)

            async def limited():
                async with semaphore.semaphore:
                    return 1
        """
    
    cdef Semaphore c_get_semaphore(self):
        return self.dummy if self.use_dummy else self.semaphores[current_thread()]

    async def __aenter__(self):
        await self.c_get_semaphore().c_acquire()

    async def __aexit__(self, *args):
        self.c_get_semaphore().c_release()



from a_sync._typing import *
import async_property as ap
from _typeshed import Incomplete
from a_sync import exceptions as exceptions
from a_sync.a_sync import config as config
from a_sync.a_sync._descriptor import ASyncDescriptor as ASyncDescriptor
from a_sync.a_sync.function import (
    ASyncFunction as ASyncFunction,
    ASyncFunctionAsyncDefault as ASyncFunctionAsyncDefault,
    ASyncFunctionSyncDefault as ASyncFunctionSyncDefault,
)
from a_sync.a_sync.method import (
    ASyncBoundMethodAsyncDefault as ASyncBoundMethodAsyncDefault,
    ASyncMethodDescriptorAsyncDefault as ASyncMethodDescriptorAsyncDefault,
)
from a_sync.task import TaskMapping as TaskMapping
from collections.abc import Generator
from typing import Any
from typing_extensions import Unpack

logger: Incomplete

class _ASyncPropertyDescriptorBase(ASyncDescriptor[I, Tuple[()], T]):
    """Base class for creating asynchronous properties.

    This class provides the foundation for defining properties that can be accessed
    both synchronously and asynchronously. It includes utility methods for common
    operations such as `any`, `all`, `min`, `max`, and `sum`.
    """

    any: ASyncFunction[AnyIterable[I], bool]
    all: ASyncFunction[AnyIterable[I], bool]
    min: ASyncFunction[AnyIterable[I], T]
    max: ASyncFunction[AnyIterable[I], T]
    sum: ASyncFunction[AnyIterable[I], T]
    hidden_method_descriptor: HiddenMethodDescriptor[T]
    __wrapped__: Callable[[I], T]
    hidden_method_name: Incomplete
    def __init__(
        self,
        _fget: AsyncGetterFunction[I, T],
        field_name: Optional[str] = None,
        **modifiers: Unpack[ModifierKwargs]
    ) -> None:
        """Initializes the _ASyncPropertyDescriptorBase.

        Args:
            _fget: The function to be wrapped.
            field_name: Optional name for the field. If not provided, the function's name will be used.
            **modifiers: Additional modifier arguments.
        """

    @overload
    def __get__(self, instance: None, owner: Type[I]) -> Self: ...
    @overload
    def __get__(self, instance: I, owner: Type[I]) -> Awaitable[T]: ...
    async def get(self, instance: I, owner: Optional[Type[I]] = None) -> T:
        """Asynchronously retrieves the property value.

        Args:
            instance: The instance from which the property is accessed.
            owner: The owner class of the property.

        Returns:
            The property value.
        """

    def map(
        self,
        instances: AnyIterable[I],
        owner: Optional[Type[I]] = None,
        concurrency: Optional[int] = None,
        name: str = "",
    ) -> TaskMapping[I, T]:
        """Maps the property across multiple instances.

        Args:
            instances: An iterable of instances.
            owner: The owner class of the property.
            concurrency: Optional concurrency limit.
            name: Optional name for the task mapping.

        Returns:
            A TaskMapping object.
        """

class ASyncPropertyDescriptor(
    _ASyncPropertyDescriptorBase[I, T], ap.base.AsyncPropertyDescriptor
):
    """Descriptor class for asynchronous properties."""

class property(ASyncPropertyDescriptor[I, T]):
    """Descriptor for defining properties that can be accessed both synchronously and asynchronously."""

class ASyncPropertyDescriptorSyncDefault(property[I, T]):
    """
    A variant of :class:`~ASyncPropertyDescriptor` that defaults to synchronous behavior.

    This class is used when the property is primarily intended to be accessed
    synchronously but can also be used asynchronously if needed.
    """

    default: str
    any: ASyncFunctionSyncDefault[AnyIterable[I], bool]
    all: ASyncFunctionSyncDefault[AnyIterable[I], bool]
    min: ASyncFunctionSyncDefault[AnyIterable[I], T]
    max: ASyncFunctionSyncDefault[AnyIterable[I], T]
    sum: ASyncFunctionSyncDefault[AnyIterable[I], T]
    @overload
    def __get__(self, instance: None, owner: Type[I]) -> Self: ...
    @overload
    def __get__(self, instance: I, owner: Type[I]) -> T: ...
    __get__: Incomplete

class ASyncPropertyDescriptorAsyncDefault(property[I, T]):
    """
    A variant of :class:`~ASyncPropertyDescriptor` that defaults to asynchronous behavior.

    This class is used when the property is primarily intended to be accessed
    asynchronously but can also be used synchronously if needed.
    """

    default: str
    any: ASyncFunctionAsyncDefault[AnyIterable[I], bool]
    all: ASyncFunctionAsyncDefault[AnyIterable[I], bool]
    min: ASyncFunctionAsyncDefault[AnyIterable[I], T]
    max: ASyncFunctionAsyncDefault[AnyIterable[I], T]
    sum: ASyncFunctionAsyncDefault[AnyIterable[I], T]

ASyncPropertyDecorator = Callable[[AnyGetterFunction[I, T]], property[I, T]]
ASyncPropertyDecoratorSyncDefault = Callable[
    [AnyGetterFunction[I, T]], ASyncPropertyDescriptorSyncDefault[I, T]
]
ASyncPropertyDecoratorAsyncDefault = Callable[
    [AnyGetterFunction[I, T]], ASyncPropertyDescriptorAsyncDefault[I, T]
]

@overload
def a_sync_property(
    func: Literal[None] = None, **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDecorator[I, T]: ...
@overload
def a_sync_property(
    func: AnyGetterFunction[I, T], **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDescriptor[I, T]: ...
@overload
def a_sync_property(
    func: Literal[None], default: Literal["sync"], **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDecoratorSyncDefault[I, T]: ...
@overload
def a_sync_property(
    func: Literal[None], default: Literal["sync"], **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDecoratorSyncDefault[I, T]: ...
@overload
def a_sync_property(
    func: Literal[None], default: Literal["async"], **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDecoratorAsyncDefault[I, T]: ...
@overload
def a_sync_property(
    func: Literal[None], default: DefaultMode = ..., **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDecorator[I, T]: ...
@overload
def a_sync_property(
    default: Literal["sync"], **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDecoratorSyncDefault[I, T]: ...
@overload
def a_sync_property(
    default: Literal["async"], **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDecoratorAsyncDefault[I, T]: ...
@overload
def a_sync_property(
    func: AnyGetterFunction[I, T],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDescriptorSyncDefault[I, T]: ...
@overload
def a_sync_property(
    func: AnyGetterFunction[I, T],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDescriptorAsyncDefault[I, T]: ...
@overload
def a_sync_property(
    func: AnyGetterFunction[I, T],
    default: DefaultMode = ...,
    **modifiers: Unpack[ModifierKwargs]
) -> ASyncPropertyDescriptor[I, T]: ...

class ASyncCachedPropertyDescriptor(
    _ASyncPropertyDescriptorBase[I, T], ap.cached.AsyncCachedPropertyDescriptor
):
    """
    A descriptor class for dual-function sync/async cached properties.

    This class extends the API of ASyncPropertyDescriptor to provide
    caching functionality, storing the computed value after the first access.
    """

    def __init__(
        self,
        _fget: AsyncGetterFunction[I, T],
        _fset: Incomplete | None = None,
        _fdel: Incomplete | None = None,
        field_name: Incomplete | None = None,
        **modifiers: Unpack[ModifierKwargs]
    ) -> None:
        """Initializes the ASyncCachedPropertyDescriptor.

        Args:
            _fget: The function to be wrapped.
            _fset: Optional setter function for the property.
            _fdel: Optional deleter function for the property.
            field_name: Optional name for the field. If not provided, the function's name will be used.
            **modifiers: Additional modifier arguments.
        """

    def get_lock(self, instance: I) -> asyncio.Task[T]:
        """Retrieves the lock for the property.

        Args:
            instance: The instance from which the property is accessed.

        Returns:
            An asyncio Task representing the lock.
        """

    def pop_lock(self, instance: I) -> None:
        """Removes the lock for the property.

        Args:
            instance: The instance from which the property is accessed.
        """

    def get_loader(self, instance: I) -> Callable[[], T]:
        """Retrieves the loader function for the property.

        Args:
            instance: The instance from which the property is accessed.

        Returns:
            A callable that loads the property value.
        """

class cached_property(ASyncCachedPropertyDescriptor[I, T]):
    """Descriptor for defining cached properties that can be accessed both synchronously and asynchronously."""

class ASyncCachedPropertyDescriptorSyncDefault(cached_property[I, T]):
    """
    A variant of :class:`~ASyncCachedPropertyDescriptor` that defaults to synchronous behavior.

    This class is used for cached properties that are primarily intended to be
    accessed synchronously but can also be used asynchronously if needed.
    """

    default: Literal["sync"]
    @overload
    def __get__(self, instance: None, owner: Type[I]) -> Self: ...
    @overload
    def __get__(self, instance: I, owner: Type[I]) -> T: ...
    __get__: Incomplete

class ASyncCachedPropertyDescriptorAsyncDefault(cached_property[I, T]):
    """
    A variant of :class:`~ASyncCachedPropertyDescriptor` that defaults to asynchronous behavior.

    This class is used for cached properties that are primarily intended to be
    accessed asynchronously but can also be used synchronously if needed.
    """

    default: Literal["async"]

ASyncCachedPropertyDecorator = Callable[
    [AnyGetterFunction[I, T]], cached_property[I, T]
]
ASyncCachedPropertyDecoratorSyncDefault = Callable[
    [AnyGetterFunction[I, T]], ASyncCachedPropertyDescriptorSyncDefault[I, T]
]
ASyncCachedPropertyDecoratorAsyncDefault = Callable[
    [AnyGetterFunction[I, T]], ASyncCachedPropertyDescriptorAsyncDefault[I, T]
]

@overload
def a_sync_cached_property(
    func: Literal[None] = None, **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDecorator[I, T]: ...
@overload
def a_sync_cached_property(
    func: AnyGetterFunction[I, T], **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDescriptor[I, T]: ...
@overload
def a_sync_cached_property(
    func: Literal[None], default: Literal["sync"], **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDecoratorSyncDefault[I, T]: ...
@overload
def a_sync_cached_property(
    func: Literal[None], default: Literal["async"], **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDecoratorAsyncDefault[I, T]: ...
@overload
def a_sync_cached_property(
    func: Literal[None], default: DefaultMode, **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDecorator[I, T]: ...
@overload
def a_sync_cached_property(
    default: Literal["sync"], **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDecoratorSyncDefault[I, T]: ...
@overload
def a_sync_cached_property(
    default: Literal["async"], **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDecoratorAsyncDefault[I, T]: ...
@overload
def a_sync_cached_property(
    func: AnyGetterFunction[I, T],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDescriptorSyncDefault[I, T]: ...
@overload
def a_sync_cached_property(
    func: AnyGetterFunction[I, T],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDescriptorAsyncDefault[I, T]: ...
@overload
def a_sync_cached_property(
    func: AnyGetterFunction[I, T],
    default: DefaultMode = ...,
    **modifiers: Unpack[ModifierKwargs]
) -> ASyncCachedPropertyDescriptor[I, T]: ...

class HiddenMethod(ASyncBoundMethodAsyncDefault[I, Tuple[()], T]):
    """Represents a hidden method for asynchronous properties.

    This class is used internally to manage hidden methods associated with
    asynchronous properties.
    """

    def __init__(
        self,
        instance: I,
        unbound: AnyFn[Concatenate[I, P], T],
        async_def: bool,
        field_name: str,
        **modifiers: Unpack[ModifierKwargs]
    ) -> None:
        """Initializes the HiddenMethod.

        Args:
            instance: The instance to which the method is bound.
            unbound: The unbound function to be wrapped.
            async_def: Indicates if the method is asynchronous.
            field_name: The name of the field associated with the method.
            **modifiers: Additional modifier arguments.
        """

    def __await__(self) -> Generator[Any, None, T]:
        """Returns an awaitable for the method."""

class HiddenMethodDescriptor(ASyncMethodDescriptorAsyncDefault[I, Tuple[()], T]):
    """Descriptor for hidden methods associated with asynchronous properties.

    This class is used internally to manage hidden methods associated with
    asynchronous properties.
    """

    __doc__: Incomplete
    def __init__(
        self,
        _fget: AnyFn[Concatenate[I, P], Awaitable[T]],
        field_name: Optional[str] = None,
        **modifiers: Unpack[ModifierKwargs]
    ) -> None:
        """
        Initialize the HiddenMethodDescriptor.

        Args:
            _fget: The function to be wrapped.
            field_name: Optional name for the field. If not provided, the function's name will be used.
            **modifiers: Additional modifier arguments.

        Raises:
            ValueError: If _fget is not callable.
        """

    def __get__(self, instance: I, owner: Type[I]) -> HiddenMethod[I, T]:
        """Retrieves the hidden method for the property.

        Args:
            instance: The instance from which the method is accessed.
            owner: The owner class of the method.

        Returns:
            The hidden method.
        """



import functools
import inspect
import logging
from contextlib import suppress

from a_sync import exceptions
from a_sync._typing import *
from a_sync.a_sync._flags cimport negate_if_necessary
from a_sync.a_sync.abstract import ASyncABC
from a_sync.a_sync.flags import VIABLE_FLAGS


logger = logging.getLogger(__name__)

cdef object c_logger = logger


class ASyncGenericBase(ASyncABC):
    """
    Base class for creating dual-function sync/async-capable classes without writing all your code twice.

    This class, via its inherited metaclass :class:`~ASyncMeta', provides the foundation for creating hybrid sync/async classes. It allows methods
    and properties to be defined once and used in both synchronous and asynchronous contexts.

    The class uses the :func:`a_sync` decorator internally to create dual-mode methods and properties.
    Subclasses should define their methods as coroutines (using `async def`) where possible, and
    use the `@a_sync.property` or `@a_sync.cached_property` decorators for properties that need to support both modes.

    Example:
        ```python
        class MyClass(ASyncGenericBase):
            def __init__(self, sync: bool):
                self.sync = sync

            @a_sync.property
            async def my_property(self):
                return await some_async_operation()

            @a_sync
            async def my_method(self):
                return await another_async_operation()

        # Synchronous usage
        obj = MyClass(sync=True)
        sync_result = obj.my_property
        sync_method_result = obj.my_method()

        # Asynchronous usage
        obj = MyClass(sync=False)
        async_result = await obj.my_property
        async_method_result = await obj.my_method()
        ```

    Note:
        When subclassing, be aware that all async methods and properties will be
        automatically wrapped to support both sync and async calls. This allows for
        seamless usage in different contexts without changing the underlying implementation.
    """

    def __init__(self):
        cdef str cls_name
        if type(self) is ASyncGenericBase:
            cls_name = type(self).__name__
            raise NotImplementedError(
                f"You should not create instances of `{cls_name}` directly, you should subclass `ASyncGenericBase` instead."
            )
        ASyncABC.__init__(self)

    @functools.cached_property
    def __a_sync_flag_name__(self) -> str:
        # TODO: cythonize this cache
        c_logger.debug("checking a_sync flag for %s", self)
        try:
            flag = _get_a_sync_flag_name_from_signature(type(self))
        except exceptions.ASyncFlagException:
            # We can't get the flag name from the __init__ signature,
            # but maybe the implementation sets the flag somewhere else.
            # Let's check the instance's atributes
            c_logger.debug(
                "unable to find flag name using `%s.__init__` signature, checking for flag attributes defined on %s",
                self.__class__.__name__,
                self,
            )
            present_flags = [flag for flag in VIABLE_FLAGS if hasattr(self, flag)]
            if not present_flags:
                raise exceptions.NoFlagsFound(self) from None
            if len(present_flags) > 1:
                raise exceptions.TooManyFlags(self, present_flags) from None
            flag = present_flags[0]
        if not isinstance(flag, str):
            raise exceptions.InvalidFlag(flag)
        return flag

    @functools.cached_property
    def __a_sync_flag_value__(self) -> bint:
        # TODO: cythonize this cache
        """If you wish to be able to hotswap default modes, just duplicate this def as a non-cached property."""
        if c_logger.isEnabledFor(logging.DEBUG):
            flag = self.__a_sync_flag_name__
            flag_value = getattr(self, flag)
            c_logger._log(logging.DEBUG, "`%s.%s` is currently %s", self, flag, flag_value)

        else:
            flag_value = getattr(self, self.__a_sync_flag_name__)

        if not isinstance(flag_value, bool):
            raise exceptions.InvalidFlagValue(flag, flag_value)

        return flag_value

    @classmethod  # type: ignore [misc]
    def __a_sync_default_mode__(cls) -> bint:  # type: ignore [override]
        cdef object flag
        cdef bint flag_value
        if not c_logger.isEnabledFor(logging.DEBUG):
            # we can optimize this if we dont need to log `flag` and the return value
            try:
                flag = _get_a_sync_flag_name_from_signature(cls)
                flag_value = _a_sync_flag_default_value_from_signature(cls)
            except exceptions.NoFlagsFound:
                flag = _get_a_sync_flag_name_from_class_def(cls)
                flag_value = _get_a_sync_flag_value_from_class_def(cls, flag)
            try:
                return negate_if_necessary(flag, flag_value)  # type: ignore [arg-type]
            except TypeError as e:
                raise exceptions.InvalidFlagValue(flag, flag_value) from e.__cause__

        # we need an extra var so we can log it
        cdef bint sync
        
        try:
            flag = _get_a_sync_flag_name_from_signature(cls)
            flag_value = _a_sync_flag_default_value_from_signature(cls)
        except exceptions.NoFlagsFound:
            flag = _get_a_sync_flag_name_from_class_def(cls)
            flag_value = _get_a_sync_flag_value_from_class_def(cls, flag)
        
        try:
            sync = negate_if_necessary(flag, flag_value)  # type: ignore [arg-type]
        except TypeError as e:
            raise exceptions.InvalidFlagValue(flag, flag_value) from e.__cause__

        c_logger._log(
            logging.DEBUG,
            "`%s.%s` indicates default mode is %ssynchronous",
            cls,
            flag,
            "a" if sync is False else "",
        )
        return sync



cdef str _get_a_sync_flag_name_from_class_def(object cls):
    c_logger.debug("Searching for flags defined on %s", cls)
    try:
        return _parse_flag_name_from_list(cls, cls.__dict__)  # type: ignore [arg-type]
        # idk why __dict__ doesn't type check as a dict
    except exceptions.NoFlagsFound:
        for base in cls.__bases__:
            with suppress(exceptions.NoFlagsFound):
                return _parse_flag_name_from_list(cls, base.__dict__)  # type: ignore [arg-type]
                # idk why __dict__ doesn't type check as a dict
    raise exceptions.NoFlagsFound(cls, list(cls.__dict__.keys()))


cdef bint _a_sync_flag_default_value_from_signature(object cls):
    cdef object signature = inspect.signature(cls.__init__)
    if not c_logger.isEnabledFor(logging.DEBUG):
        # we can optimize this much better
        return signature.parameters[_get_a_sync_flag_name_from_signature(cls)].default
    
    c_logger._log(
        logging.DEBUG, "checking `__init__` signature for default %s a_sync flag value", cls
    )
    cdef str flag = _get_a_sync_flag_name_from_signature(cls)
    cdef object flag_value = signature.parameters[flag].default
    if flag_value is inspect._empty:  # type: ignore [attr-defined]
        raise NotImplementedError(
            "The implementation for 'cls' uses an arg to specify sync mode, instead of a kwarg. We are unable to proceed. I suppose we can extend the code to accept positional arg flags if necessary"
        )
    c_logger._log(logging.DEBUG, "%s defines %s, default value %s", cls, flag, flag_value)
    return flag_value


cdef str _get_a_sync_flag_name_from_signature(object cls):
    if cls.__name__ == "ASyncGenericBase":
        c_logger.debug(
            "There are no flags defined on the base class, this is expected. Skipping."
        )
        return None

    # if we fail this one there's no need to check again
    if not c_logger.isEnabledFor(logging.DEBUG):
        # we can also skip assigning params to a var
        return _parse_flag_name_from_list(cls, inspect.signature(cls.__init__).parameters)

    c_logger._log(logging.DEBUG, "Searching for flags defined on %s.__init__", cls)
    cdef object parameters = inspect.signature(cls.__init__).parameters
    c_logger._log(logging.DEBUG, "parameters: %s", parameters)
    return _parse_flag_name_from_list(cls, parameters)


cdef str _parse_flag_name_from_list(object cls, object items):
    cdef str flag
    cdef list[str] present_flags = [flag for flag in VIABLE_FLAGS if flag in items]
    if not present_flags:
        c_logger.debug("There are too many flags defined on %s", cls)
        raise exceptions.NoFlagsFound(cls, items.keys())
    if len(present_flags) > 1:
        c_logger.debug("There are too many flags defined on %s", cls)
        raise exceptions.TooManyFlags(cls, present_flags)
    if c_logger.isEnabledFor(logging.DEBUG):
        flag = present_flags[0]
        c_logger._log(logging.DEBUG, "found flag %s", flag)
        return flag
    return present_flags[0]


cdef bint _get_a_sync_flag_value_from_class_def(object cls, str flag):
    cdef object spec
    for spec in [cls, *cls.__bases__]:
        if flag in spec.__dict__:
            return spec.__dict__[flag]
    raise exceptions.FlagNotDefined(cls, flag)



from a_sync._typing import *
import asyncio
from _typeshed import Incomplete
from a_sync.primitives.locks.semaphore import Semaphore as Semaphore
from functools import cached_property as cached_property
from typing import TypeVar

logger: Incomplete

class Priority(Protocol):
    def __lt__(self, other) -> bool: ...

PT = TypeVar("PT", bound=Priority)
CM = TypeVar("CM", bound="_AbstractPrioritySemaphoreContextManager[Priority]")

class _AbstractPrioritySemaphore(Semaphore, Generic[PT, CM]):
    """
    A semaphore that allows prioritization of waiters.

    This semaphore manages waiters with associated priorities, ensuring that waiters with higher
    priorities are processed before those with lower priorities. Subclasses must define the
    `_top_priority` attribute to specify the default top priority behavior.

    The `_context_manager_class` attribute should specify the class used for managing semaphore contexts.

    See Also:
        :class:`PrioritySemaphore` for an implementation using numeric priorities.
    """

    name: Optional[str]
    def __init__(self, value: int = 1, *, name: Optional[str] = None) -> None:
        """Initializes the priority semaphore.

        Args:
            value: The initial capacity of the semaphore.
            name: An optional name for the semaphore, used for debugging.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5, name="test_semaphore")
        """

    async def __aenter__(self) -> None:
        """Enters the semaphore context, acquiring it with the top priority.

        This method is part of the asynchronous context management protocol.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> async with semaphore:
            ...     await do_stuff()
        """

    async def __aexit__(self, *_) -> None:
        """Exits the semaphore context, releasing it with the top priority.

        This method is part of the asynchronous context management protocol.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> async with semaphore:
            ...     await do_stuff()
        """

    async def acquire(self) -> Literal[True]:
        """Acquires the semaphore with the top priority.

        This method overrides :meth:`Semaphore.acquire` to handle priority-based logic.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> await semaphore.acquire()
        """

    def __getitem__(
        self, priority: Optional[PT]
    ) -> _AbstractPrioritySemaphoreContextManager[PT]:
        """Gets the context manager for a given priority.

        Args:
            priority: The priority for which to get the context manager. If None, uses the top priority.

        Returns:
            The context manager associated with the given priority.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> context_manager = semaphore[priority]
        """

    def locked(self) -> bool:
        """Checks if the semaphore is locked.

        Returns:
            True if the semaphore cannot be acquired immediately, False otherwise.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> semaphore.locked()
        """

class _AbstractPrioritySemaphoreContextManager(Semaphore, Generic[PT]):
    """
    A context manager for priority semaphore waiters.

    This context manager is associated with a specific priority and handles
    the acquisition and release of the semaphore for waiters with that priority.
    """

    def __init__(
        self,
        parent: _AbstractPrioritySemaphore,
        priority: PT,
        name: Optional[str] = None,
    ) -> None:
        """Initializes the context manager for a specific priority.

        Args:
            parent: The parent semaphore.
            priority: The priority associated with this context manager.
            name: An optional name for the context manager, used for debugging.

        Examples:
            >>> parent_semaphore = _AbstractPrioritySemaphore(5)
            >>> context_manager = _AbstractPrioritySemaphoreContextManager(parent_semaphore, priority=1)
        """

    def __lt__(self, other) -> bool:
        """Compares this context manager with another based on priority.

        Args:
            other: The other context manager to compare with.

        Returns:
            True if this context manager has a lower priority than the other, False otherwise.

        Raises:
            TypeError: If the other object is not of the same type.

        Examples:
            >>> cm1 = _AbstractPrioritySemaphoreContextManager(parent, priority=1)
            >>> cm2 = _AbstractPrioritySemaphoreContextManager(parent, priority=2)
            >>> cm1 < cm2
        """

    @cached_property
    def loop(self) -> asyncio.AbstractEventLoop:
        """Gets the event loop associated with this context manager."""

    @property
    def waiters(self) -> Deque[asyncio.Future]:
        """Gets the deque of waiters for this context manager."""

    async def acquire(self) -> Literal[True]:
        """Acquires the semaphore for this context manager.

        If the internal counter is larger than zero on entry,
        decrement it by one and return True immediately. If it is
        zero on entry, block, waiting until some other coroutine has
        called release() to make it larger than 0, and then return
        True.

        This method overrides :meth:`Semaphore.acquire` to handle priority-based logic.

        Examples:
            >>> context_manager = _AbstractPrioritySemaphoreContextManager(parent, priority=1)
            >>> await context_manager.acquire()
        """

    def release(self) -> None:
        """Releases the semaphore for this context manager.

        This method overrides :meth:`Semaphore.release` to handle priority-based logic.

        Examples:
            >>> context_manager = _AbstractPrioritySemaphoreContextManager(parent, priority=1)
            >>> context_manager.release()
        """

class _PrioritySemaphoreContextManager(
    _AbstractPrioritySemaphoreContextManager[Numeric]
):
    """Context manager for numeric priority semaphores."""

class PrioritySemaphore(
    _AbstractPrioritySemaphore[Numeric, _PrioritySemaphoreContextManager]
):
    """Semaphore that uses numeric priorities for waiters.

    This class extends :class:`_AbstractPrioritySemaphore` and provides a concrete implementation
    using numeric priorities. The `_context_manager_class` is set to :class:`_PrioritySemaphoreContextManager`,
    and the `_top_priority` is set to -1, which is the highest priority.

    Examples:
        The primary way to use this semaphore is by specifying a priority.

        >>> priority_semaphore = PrioritySemaphore(10)
        >>> async with priority_semaphore[priority]:
        ...     await do_stuff()

        You can also enter and exit this semaphore without specifying a priority, and it will use the top priority by default:

        >>> priority_semaphore = PrioritySemaphore(10)
        >>> async with priority_semaphore:
        ...     await do_stuff()

    See Also:
        :class:`_AbstractPrioritySemaphore` for the base class implementation.
    """



"""
This module provides utility functions for handling keyword arguments related to synchronous and asynchronous flags.
"""

from libc.stdint cimport uint8_t

from a_sync import exceptions
from a_sync.a_sync._flags cimport negate_if_necessary
from a_sync.a_sync.flags import VIABLE_FLAGS


def get_flag_name_legacy(dict kwargs) -> str:
    return get_flag_name(kwargs)


cdef str get_flag_name(dict kwargs):
    """
    Get the name of the flag present in the kwargs.

    Args:
        kwargs: A dictionary of keyword arguments.

    Returns:
        The name of the flag if present, an empty string otherwise.

    Raises:
        :class:`exceptions.TooManyFlags`: If more than one flag is present in the kwargs,
        the exception includes the message "kwargs" and the list of present flags.

    Examples:
        >>> get_flag_name({'sync': True})
        'sync'

        >>> get_flag_name({'asynchronous': False})
        'asynchronous'

        >>> get_flag_name({})
        ''
    """
    cdef list present_flags = [flag for flag in VIABLE_FLAGS if flag in kwargs]
    cdef uint8_t flags_count = len(present_flags)
    if flags_count == 0:
        return ""  # indicates no flag is present
    elif flags_count == 1:
        return present_flags[0]
    raise exceptions.TooManyFlags("kwargs", present_flags)


cdef bint is_sync(str flag, dict kwargs, bint pop_flag):
    """
    Determine if the operation should be synchronous based on the flag value.

    Args:
        flag (str): The name of the flag to check.
        kwargs (dict): A dictionary of keyword arguments.
        pop_flag (bool, optional): Whether to remove the flag from kwargs. Defaults to False.

    See Also:
        :func:`get_flag_name`: Retrieves the name of the flag present in the kwargs.
    """
    if pop_flag:
        # NOTE: we should techincally raise InvalidFlagValue here but I dont want
        #       to set flag_value to a var and it will raise a TypeError anyway.
        return negate_if_necessary(flag, kwargs.pop(flag))
    else:
        try:
            return negate_if_necessary(flag, kwargs[flag])
        except TypeError as e:
            raise exceptions.InvalidFlagValue(flag, kwargs[flag]) from e.__cause__


"""
This module provides configuration options and default settings for the a_sync library.
It includes functionality for setting up executors, defining default modifiers,
and handling environment variable configurations.

Environment Variables:
    :obj:`~A_SYNC_EXECUTOR_TYPE`: Specifies the type of executor to use. Valid values are
        strings that start with 'p' for :class:`~concurrent.futures.ProcessPoolExecutor` 
        (e.g., 'processes') or 't' for :class:`~concurrent.futures.ThreadPoolExecutor` 
        (e.g., 'threads'). Defaults to 'threads'.
    :obj:`~A_SYNC_EXECUTOR_VALUE`: Specifies the number of workers for the executor.
        Defaults to 8.
    :obj:`~A_SYNC_DEFAULT_MODE`: Sets the default mode for a_sync functions if not specified.
    :obj:`~A_SYNC_CACHE_TYPE`: Sets the default cache type. If not specified, defaults to None.
    :obj:`~A_SYNC_CACHE_TYPED`: Boolean flag to determine if cache keys should consider types.
    :obj:`~A_SYNC_RAM_CACHE_MAXSIZE`: Sets the maximum size for the RAM cache. Defaults to -1.
    :obj:`~A_SYNC_RAM_CACHE_TTL`: Sets the time-to-live for cache entries. If not specified,
        defaults to 0, which is then checked against `null_modifiers["ram_cache_ttl"]` 
        to potentially set it to `None`, meaning cache entries do not expire by default.
    :obj:`~A_SYNC_RUNS_PER_MINUTE`: Sets the rate limit for function execution.
    :obj:`~A_SYNC_SEMAPHORE`: Sets the semaphore limit for function execution.

Examples:
    To set the executor type to use threads with 4 workers, set the environment variables:
    
    .. code-block:: bash

        export A_SYNC_EXECUTOR_TYPE=threads
        export A_SYNC_EXECUTOR_VALUE=4

    To configure caching with a maximum size of 100 and a TTL of 60 seconds:

    .. code-block:: bash

        export A_SYNC_CACHE_TYPE=memory
        export A_SYNC_RAM_CACHE_MAXSIZE=100
        export A_SYNC_RAM_CACHE_TTL=60

TODO: explain how and where these values are used

See Also:
    - :mod:`concurrent.futures`: For more details on executors.
    - :mod:`functools`: For caching mechanisms.
"""

import functools
import os
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from concurrent.futures._base import Executor

from a_sync._typing import *

EXECUTOR_TYPE = os.environ.get("A_SYNC_EXECUTOR_TYPE", "threads")
"""Specifies the type of executor to use.

Valid values are strings that start with 'p' for :class:`~concurrent.futures.ProcessPoolExecutor`
(e.g., 'processes') or 't' for :class:`~concurrent.futures.ThreadPoolExecutor` (e.g., 'threads').
Defaults to 'threads'.
"""

EXECUTOR_VALUE = int(os.environ.get("A_SYNC_EXECUTOR_VALUE", 8))
"""Specifies the number of workers for the executor. Defaults to 8."""


@functools.lru_cache(maxsize=1)
def get_default_executor() -> Executor:
    """Get the default executor based on the EXECUTOR_TYPE environment variable.

    Returns:
        An instance of either :class:`~concurrent.futures.ProcessPoolExecutor`
        or :class:`~concurrent.futures.ThreadPoolExecutor`.

    Raises:
        ValueError: If an invalid EXECUTOR_TYPE is specified. Valid values are
        strings that start with 'p' for :class:`~concurrent.futures.ProcessPoolExecutor`
        or 't' for :class:`~concurrent.futures.ThreadPoolExecutor`.

    Examples:
        >>> import os
        >>> os.environ["A_SYNC_EXECUTOR_TYPE"] = "threads"
        >>> executor = get_default_executor()
        >>> isinstance(executor, ThreadPoolExecutor)
        True

        >>> os.environ["A_SYNC_EXECUTOR_TYPE"] = "processes"
        >>> executor = get_default_executor()
        >>> isinstance(executor, ProcessPoolExecutor)
        True
    """
    if EXECUTOR_TYPE.lower().startswith("p"):  # p, P, proc, Processes, etc
        return ProcessPoolExecutor(EXECUTOR_VALUE)
    elif EXECUTOR_TYPE.lower().startswith("t"):  # t, T, thread, THREADS, etc
        return ThreadPoolExecutor(EXECUTOR_VALUE)
    raise ValueError(
        "Invalid value for A_SYNC_EXECUTOR_TYPE. Please use 'threads' or 'processes'."
    )


default_sync_executor = get_default_executor()

null_modifiers = ModifierKwargs(
    default=None,
    cache_type=None,
    cache_typed=False,
    ram_cache_maxsize=-1,
    ram_cache_ttl=None,
    runs_per_minute=None,
    semaphore=None,
    executor=default_sync_executor,
)

#####################
# Default Modifiers #
#####################

# User configurable default modifiers to be applied to any a_sync decorated function if you do not specify kwarg values for each modifier.

DEFAULT_MODE: DefaultMode = os.environ.get("A_SYNC_DEFAULT_MODE")  # type: ignore [assignment]
"""Sets the default mode for a_sync functions if not specified."""

CACHE_TYPE: CacheType = (
    typ
    if (typ := os.environ.get("A_SYNC_CACHE_TYPE", "").lower())
    else null_modifiers["cache_type"]
)
"""Sets the default cache type. If not specified, defaults to None."""

CACHE_TYPED = bool(os.environ.get("A_SYNC_CACHE_TYPED"))
"""Boolean flag to determine if cache keys should consider types."""

RAM_CACHE_MAXSIZE = int(os.environ.get("A_SYNC_RAM_CACHE_MAXSIZE", -1))
"""
Sets the maximum size for the RAM cache. Defaults to -1.
# TODO: explain what -1 does
"""

RAM_CACHE_TTL = (
    ttl
    if (ttl := float(os.environ.get("A_SYNC_RAM_CACHE_TTL", 0)))
    else null_modifiers["ram_cache_ttl"]
)
"""
Sets the time-to-live for cache entries. If not specified, defaults to 0, which is then checked against
`null_modifiers["ram_cache_ttl"]`  to potentially set it to `None`, meaning cache entries do not expire
by default.
"""


RUNS_PER_MINUTE = (
    rpm
    if (rpm := int(os.environ.get("A_SYNC_RUNS_PER_MINUTE", 0)))
    else null_modifiers["runs_per_minute"]
)
"""Sets the rate limit for function execution."""

SEMAPHORE = (
    rpm
    if (rpm := int(os.environ.get("A_SYNC_SEMAPHORE", 0)))
    else null_modifiers["semaphore"]
)
"""Sets the semaphore limit for function execution."""

user_set_default_modifiers = ModifierKwargs(
    default=DEFAULT_MODE,
    cache_type=CACHE_TYPE,
    cache_typed=CACHE_TYPED,
    ram_cache_maxsize=RAM_CACHE_MAXSIZE,
    ram_cache_ttl=RAM_CACHE_TTL,
    runs_per_minute=RUNS_PER_MINUTE,
    semaphore=SEMAPHORE,
    executor=default_sync_executor,
)



"""
This module provides functionality for handling synchronous and asynchronous flags
in the ez-a-sync library.

ez-a-sync uses 'flags' to indicate whether objects or function calls will be synchronous or asynchronous.

You can use any of the provided flags, whichever makes the most sense for your use case.

:obj:`AFFIRMATIVE_FLAGS`: Set of flags indicating synchronous behavior. Currently includes "sync".

:obj:`NEGATIVE_FLAGS`: Set of flags indicating asynchronous behavior. Currently includes "asynchronous".

:obj:`VIABLE_FLAGS`: Set of all valid flags, combining both synchronous and asynchronous indicators.
"""

from a_sync import exceptions
from a_sync.a_sync.flags import AFFIRMATIVE_FLAGS, NEGATIVE_FLAGS


cdef bint negate_if_necessary(str flag, bint flag_value):
    """Negate the flag value if necessary based on the flag type.

    This function checks if the provided flag is in the set of affirmative or negative flags
    and negates the flag value accordingly. If the flag is not recognized, it raises an exception.

    Args:
        flag: The flag to check.
        flag_value: The value of the flag.

    Returns:
        The potentially negated flag value.

    Raises:
        TypeError: If the provided flag value is not of bint type.
        exceptions.InvalidFlag: If the flag is not recognized.

    Examples:
        >>> negate_if_necessary('sync', True)
        True

        >>> negate_if_necessary('asynchronous', True)
        False

    See Also:
        - :func:`validate_flag_value`: Validates that the flag value is a boolean.
    """
    if flag in AFFIRMATIVE_FLAGS:
        return flag_value
    elif flag in NEGATIVE_FLAGS:
        return not flag_value
    raise exceptions.InvalidFlag(flag)


cdef bint validate_flag_value(str flag, object flag_value):
    """
    Validate that the flag value is a boolean.

    This function ensures that the provided flag value is of type boolean. If not, it raises an exception.

    Args:
        flag: The flag being validated.
        flag_value: The value to validate.

    Returns:
        The validated flag value.

    Raises:
        exceptions.InvalidFlagValue: If the flag value is not a boolean.

    Examples:
        >>> validate_flag_value('sync', True)
        True

    See Also:
        - :func:`negate_if_necessary`: Negates the flag value if necessary based on the flag type.
    """
    if not isinstance(flag_value, bool):
        raise exceptions.InvalidFlagValue(flag, flag_value)
    return flag_value



import pytest
import asyncio

from a_sync.task import (
    TaskMapping,
    TaskMappingKeys,
    TaskMappingValues,
    TaskMappingItems,
    _EmptySequenceError,
)


views = [TaskMappingKeys, TaskMappingValues, TaskMappingItems]


async def sample_task(key):
    await asyncio.sleep(0.1)
    return f"result for {key}"


@pytest.mark.parametrize(
    "view_class, expected",
    [
        (TaskMappingKeys, ["key1", "key2"]),
        (TaskMappingValues, ["result for key1", "result for key2"]),
        (TaskMappingItems, [("key1", "result for key1"), ("key2", "result for key2")]),
    ],
)
def test_taskmapping_views_sync(view_class, expected):
    tasks = TaskMapping(sample_task, ["key1", "key2"])
    view = view_class(expected, tasks)
    results = list(view)
    assert results == expected


@pytest.mark.asyncio_cooperative
@pytest.mark.parametrize(
    "view_class, expected",
    [
        (TaskMappingKeys, ["key1", "key2"]),
        (TaskMappingValues, ["result for key1", "result for key2"]),
        (TaskMappingItems, [("key1", "result for key1"), ("key2", "result for key2")]),
    ],
)
async def test_taskmapping_views_async(view_class, expected):
    tasks = TaskMapping(sample_task, ["key1", "key2"])
    view = view_class(expected, tasks)
    results = [item async for item in view]
    assert results == expected


@pytest.mark.parametrize(
    "view_class", [TaskMappingKeys, TaskMappingValues, TaskMappingItems]
)
def test_empty_taskmapping_views_sync(view_class):
    # sourcery skip: simplify-empty-collection-comparison
    tasks = TaskMapping(sample_task)
    view = view_class([], tasks)
    results = list(view)
    assert results == []


@pytest.mark.asyncio_cooperative
@pytest.mark.parametrize(
    "view_class", [TaskMappingKeys, TaskMappingValues, TaskMappingItems]
)
async def test_empty_taskmapping_views_async(view_class):
    # sourcery skip: simplify-empty-collection-comparison
    tasks = TaskMapping(sample_task)
    view = view_class([], tasks)
    results = [item async for item in view]
    assert results == []


""" # NOTE not sure why this fails but we can get to it later
@pytest.mark.parametrize("view_class", views)
def test_empty_taskmapping_views_broken_mapping_sync(view_class):
    tasks = TaskMapping(sample_task, [])
    view = view_class([], tasks)
    with pytest.raises(_EmptySequenceError, match="\[\]"):
        results = list(view)"""


@pytest.mark.asyncio_cooperative
@pytest.mark.parametrize("view_class", views)
async def test_empty_taskmapping_views_broken_mapping_async(view_class):
    tasks = TaskMapping(sample_task, [])
    view = view_class([], tasks)
    with pytest.raises(_EmptySequenceError, match="\[\]"):
        results = [item async for item in view]


@pytest.mark.parametrize(
    "view_class, expected",
    [
        (TaskMappingKeys, ["key1"]),
        (TaskMappingValues, ["result for key1"]),
        (TaskMappingItems, [("key1", "result for key1")]),
    ],
)
def test_single_item_taskmapping_views_sync(view_class, expected):
    tasks = TaskMapping(sample_task, ["key1"])
    view = view_class(expected, tasks)
    results = list(view)
    assert results == expected


@pytest.mark.asyncio_cooperative
@pytest.mark.parametrize(
    "view_class, expected",
    [
        (TaskMappingKeys, ["key1"]),
        (TaskMappingValues, ["result for key1"]),
        (TaskMappingItems, [("key1", "result for key1")]),
    ],
)
async def test_single_item_taskmapping_views_async(view_class, expected):
    tasks = TaskMapping(sample_task, ["key1"])
    view = view_class(expected, tasks)
    results = [item async for item in view]
    assert results == expected
